;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;                                                                  ;;;;;
;;;;;           Earley-basierter Chart-parser + DRS-Generator          ;;;;;
;;;;;                                                                  ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; Datei           : earley-drt.lsp
;; letzte nderung : 22.07.96



;;; (A) Syntax/Semantik-Schnittstelle
;;; ---------------------------------

;;; Beide Funktionen lesen eine Folge von Stzen ein und berechnen fr
;;; sie eine DRS. Der Unterschied zwischen beiden Versionen besteht darin,
;;; da PARSE-TEXT1 die DRS fr die Stze erst berechnet nachdem sie alle
;;; geparst wurden, whrend PARSE-TEXT2 die Stze direkt nachdem sie geparst
;;; wurden interpretiert.

(defun parse-text1 ()
  (let (eingabe
	parse-trees)
    (loop
     (format t "~%>>> ")
     (setq eingabe (read))
     (if (member eingabe '(e ende aus))
	 (return (reduce-drs (make-instance 'drs
					    :complex-conditions parse-trees))))
     (setq parse-trees
	  (append parse-trees (parse eingabe))))))

(defun parse-text2 ()
  (let (eingabe
	(drs (make-instance 'drs)))
    (loop
     (format t "~%>>> ")
     (setq eingabe (read))
     (if (member eingabe '(e ende aus))
	 (return drs))
     (setf (complex-conditions drs)
	   (append (complex-conditions drs) (parse eingabe)))
     (reduce-drs drs))))



;;; (B) Syntax & Lexikon
;;; --------------------

(defclass syntax ()
  ((startsymbol :accessor startsymbol :initarg :s :initform 's :type symbol)
   (terminale   :accessor terminale-symbole :initarg :vt
		:initform nil :type list)
   (non-terminale :accessor nicht-terminale-symbole :initarg :vn 
		  :initform nil :type list)
   (regeln :accessor regeln :initarg :r :initform nil :type list)
   (erreichbarkeitstabelle :accessor tabelle :initarg :etab
			   :initform nil :type list)))

(defmethod print-object ((g syntax) stream)
  (format stream "~%Regeln:~%~a" (regeln g))
  (format stream "~%Tabelle:~%~a" (tabelle g)))

(defmethod vokabular ((g syntax))
  (append (nicht-terminale-symbole g) (terminale-symbole g)))

(defmethod expansionen ((g syntax) symbol)
  (apply #'append
	 (mapcar #'(lambda (x) (if (eq (first x) symbol) (list (rest x))))
		 (regeln g))))


;;; (C) Der Parser
;;; --------------

(defun parse (satz &optional (syntax *syntax*) (lexikon *lexikon*))
" Earley-basierter Recognizer fr einfache kontextfreie Syntaxen.
  kante        : (start ende kopf geschlossen offen) "
  (do* ((chart (initialisiere-chart (make-instance 'chart) satz syntax lexikon)
	       (update-chart chart kante))
	(agenda (expand (make-instance 'kante :offen `(,(startsymbol syntax)))
			syntax chart)
		(neue-kanten kante chart syntax agenda))
	(kante (first agenda) (first agenda)))
       ((endp agenda)
	;; Wenn Debugging erforderlich ist ...
	(SETQ *CHART* CHART)
	(auswerten chart satz syntax))
       ;; Wenn Debugging erforderlich ist ...       (PRINT KANTE)
       ))


(defclass chart ()
  ((aktive-kanten  :accessor akanten :initform nil :type list)
   (passive-kanten :accessor pkanten :initform nil :type list)))

(defmethod print-object ((c chart) stream)
  (print 'chart stream)
  (print (akanten c) stream)
  (print (pkanten c) stream)
  (print '-------))

(defclass kante ()
  ((start :reader   start :initarg :start :initform 0 :type integer)
   (ende  :accessor ende  :initarg :ende  :initform 0 :type integer)
   (kopf  :reader   kopf  :initarg :kopf :initform nil :type symbol)
   (geschlossen  :reader geschlossen  :initarg :geschlossen
		 :initform nil :type symbol)
   (offen :reader offen :initarg :offen :initform nil :type symbol)
   (struktur :accessor struktur :initarg :struktur :initform nil :type list)
   (fehler :accessor fehler :initarg :fehler :initform 0 :type integer)))

(defmethod aktive-kante-p ((k kante))
  (offen k))
(defmethod passive-kante-p ((k kante))
  (not (offen k)))

;; Erweiterung der generischen Funktion PRINT-OBJECT, um eine lesbare
;; Ausgabe der Kanten zu erhalten:
(defmethod print-object ((k kante) stream)
  (format stream "~%(~a ~a ~a ~a ~a ~a)"
	  (start k) (ende k) (kopf k) (geschlossen k) (offen k)
	  (struktur k) (fehler k)))

;;; Gleichheit von Kanten:
(defmethod same-kante-p ((k1 kante) (k2 kante))
  (if (or (and (aktive-kante-p k1) (aktive-kante-p k2))
	  (and (passive-kante-p k1) (passive-kante-p k2)))
      (and (= (start k1) (start k2))
	   (= (ende k1)  (ende k2))
	   (= (fehler k1) (fehler k2))
	   (eq (kopf k1)  (kopf k2))
	   (equal (offen k1) (offen k2))
	   (equal (geschlossen k1) (geschlossen k2))
	   (equal (struktur k1) (struktur k2)))))

(defmethod nicht-enthalten-p ((k kante) kanten-menge)
  (dolist (kante kanten-menge t)
     (when (same-kante-p k kante)
       (return nil))))

(defmethod initialisiere-chart ((c chart) satz (g syntax) lexikon)
  (let ((index 0))
    (dolist (wort satz c)
      (dolist (kategorie (mapcar #'first (rest (assoc wort lexikon))))
       (push (make-instance 'kante
		      :start index
		      :ende (+ index 1)
		      :kopf kategorie 
		      :geschlossen (list wort)
		      :struktur (list kategorie wort))
	     (pkanten c)))
       (incf index))))

(defmethod auswerten ((c chart) satz (g syntax))
  (let (parse-trees)
    (format t "~%  Die Chart enthlt:")
    (format t "~%  ~a ~8T aktive Kanten"
	    (length (akanten c)))
    (format t "~%  ~a ~8T passsive Kanten"
	    (length (pkanten c)))
    (dolist (kante (pkanten c) (mapc #'pprint parse-trees))
	  (when (and (= (start kante) 0)
		     (= (ende kante) (length satz))
		     (eq (kopf kante) (startsymbol g)))
		(push (struktur kante) parse-trees)))))

(defmethod update-chart ((c chart) (k kante))
  (if (aktive-kante-p k)
      (push k (akanten c))
      (push k (pkanten c)))
  c)

(defmethod neue-kanten ((k kante) (c chart) (g syntax) agenda)
  (append (rest agenda)
	  (remove-if-not
	   #'(lambda (x) (nicht-enthalten-p x (append agenda (akanten c))))
	   (if (aktive-kante-p k)
	       (append (expand k g C) (complete k c))
	       (complete k c)))))

(defmacro deleteable-p (x g)
  ;; Entweder ist X die rechte Seite einer Tilgungsregel oder 
  ;; es gibt in der Syntax G eine Tilgungsregel fr das erste
  ;; Symbol von X.
  `(or (not ,x)
       (member nil (second (assoc (first ,x) (tabelle ,g)))
	       :test #'equal)))


(defmethod expand ((k kante) (g syntax) (C CHART))
  (mapcar #'(lambda (x)
	      (make-instance 'kante
			     :start (ende k)
			     :ende  (ende k)
			     :kopf (first (offen k))
			     :offen x
			     :STRUKTUR (IF (NOT X) `(,(first (offen k)) nil))))
	  ;; Es werden nur die Regeln bercksichtigt, von deren linken Ecke
	  ;; aus der Kopf einer passiven Kante P der Chart erreichbar ist,
	  ;; fr die gilt: (ende k) = (start p).
	  (remove-if-not
	   #'(lambda (X)
	       (or (Intersection
		    (second (assoc (first  x)  (tabelle g)))
		    (pkanten-kpfe c (ende k)))
		    (deleteable-p x g)))
	  (expansionen g (first (offen k))))))

;;; Die folgende Funktion liefert eine Liste aller Kpfe der passiven Kanten
;;; der Chart, die am Punkt POSITION beginnen 
(defmethod pkanten-kpfe ((c chart) position)
  (remove-duplicates
   (mapcar #'kopf
	   (remove-if-not #'(lambda (x)
			      (= (start x) position))
			  (pkanten c)))))

(defmethod complete ((k kante) (c chart))
  (if (aktive-kante-p k)
      (mapcar #'(lambda (x) (bilde-kante k x))
	      (remove-if #'(lambda (x)
			     (not (matching-kanten-p k x)))
			 (pkanten c)))
      (mapcar #'(lambda (x) (bilde-kante x k))
	      (remove-if #'(lambda (x)
			     (not (matching-kanten-p x k)))
			 (akanten c)))))



(defmethod matching-kanten-p ((k1 kante) (k2 kante))
  (and (= (ende k1) (start k2))
       (eq (first (offen k1)) (kopf k2))))

(defmethod bilde-kante ((k1 kante) (k2 kante))
  (if (rest (offen k1))
      (make-instance 'kante
		     :start (start k1)
		     :ende (ende k2)
		     :kopf (kopf k1)
		     :geschlossen (append (geschlossen k1) (list (kopf k2)))
		     :offen (rest (offen k1))
		     :struktur (append (struktur k1) (list (struktur k2)))
		     :fehler (+ (fehler k1) (fehler k2)))
      (make-instance 'kante
		     :start (start k1)
		     :ende (ende k2)
		     :kopf (kopf k1)
		     :geschlossen (append (geschlossen k1) (list (kopf k2)))
		     :struktur (cons (kopf k1) (append (struktur k1)
						       (list (struktur k2))))
		     :fehler (+ (fehler k1) (fehler k2)))))


;;; (D) Erreichbarkeitsrelation
;;; ---------------------------

(defmethod erreichbarkeitstabelle ((g syntax))
  (setf (tabelle g)
	(mapcar #'(lambda (x) (list x (erreichbar x g)))
		(vokabular g))))

;; Reflexive + transitiver Abschlu der Relation DIREKT-ERREICHBAR
(defun erreichbar (nt-symbol syntax &optional (erreichbare-symbole `(,nt-symbol)))
   (let ((neue-symbole (set-difference
			(direkt-erreichbar nt-symbol
					   (expansionen syntax nt-symbol)
					   syntax)
			erreichbare-symbole)))
     (dolist (symbol neue-symbole (append erreichbare-symbole neue-symbole))
       (if (not (member symbol (terminale-symbole syntax)))
	   (setq erreichbare-symbole
		 (union erreichbare-symbole
			(erreichbar symbol syntax erreichbare-symbole)))))))

;; Diese Prozedur berechnet fr ein Symbol die Menge aller direkt erreichbaren
;; Symbole (linke Ecken):
(defun direkt-erreichbar (symbol expansionen syntax)
  (let (resultat)
    (dolist (x expansionen resultat)
      (unless (member (first x) resultat)
	(push (first x) resultat))
      (dolist (y (rest x))
	(when (not (member nil
			   (expansionen syntax (first x)) :test #'equal))
	      (return))
	(unless (member  y resultat)
		(push y resultat))))))


;;; (E) DRS-Generierung
;;; -------------------

(defclass drs ()
  ((universe :accessor universe :initarg :universe :initform nil :type list)
   (simple-conditions :accessor simple-conditions :initarg :simple-conditions
	       :initform nil :type list)
   (complex-conditions :accessor complex-conditions :initarg :complex-conditions
	       :initform nil :type list)
   (super-drs :accessor super :initarg :super :initform nil)))

(defmethod print-object ((d drs) stream)
  (format stream "~%#DRS<~8TU[~{~a ~}]" (universe d))
  (format stream "~%~8TSC[~{~a~}]" (simple-conditions d))
  (format stream "~%~8TCC[~{~a~}]>~%" (complex-conditions d)))
  

(defun simple-condition-p (condition)
  (or (drs-p condition)
      (not (some #'listp condition))))
		   
(defun drs-p (object)
  (typep object 'drs))


(defmethod reduce-drs ((d drs))
  (loop
   (if (not (complex-conditions d)) (return d))
   (reduce-condition d)))

(defmethod reduce-condition ((d drs))
  (do* ((condition
	 (pop (complex-conditions d))
	 (apply-rule d condition)))
       ((simple-condition-p condition)
	(push condition (simple-conditions d))
	(return d))))

(defmethod apply-rule ((d drs) condition)
  (or (apply-rc-rule d condition)
      (apply-pn-rule d condition)
      (apply-id-rule d condition)
      (apply-pro-rule d condition)
      (apply-neg-vp-rule d condition)
      (apply-vp-reduction-rule d condition)
      (apply-error-rule d condition)))

;;; Interpretationsregeln fr verschiedene Konstruktionen:
;;; ------------------------------------------------------

;; (a) Vereinfachung der VP-Struktur
;;     (entfernt S und berflssige Strukturebenen in der VP)
(defmethod apply-vp-reduction-rule ((d drs) condition)
  (let* ((vp (my-assoc 'vp condition)))
    (when (eq (first (second (second vp))) 'v)
      (setf condition
	    (cons (second condition)
		  (cons (second (second (second vp)))
			(rest (rest (second vp))))))
      condition)))

;; (b) Analyse negierter Verbalphrasen
(defmethod apply-neg-vp-rule ((d drs) condition)
  (let* ((vp (my-assoc 'vp condition))
	 new-drs)
    (when (and (eq (first (second vp)) 'aux)
	       (member (second condition) (universe d)))
      (setf new-drs (make-instance 'drs
				   :complex-conditions
				   `(,(append (butlast condition)
					      `((vp ,(fourth vp))))))
	    (super new-drs) d)
      (reduce-drs new-drs)
      new-drs)))

;; (c) Analyse von NPs mit proper nouns als Kopf
(defun pn-np (struktur)
  (let ((np (my-assoc 'np struktur)))
    (if (eq (first (second np)) 'pn)
	np)))

(defmethod apply-pn-rule ((d drs) condition)
  (let ((np (pn-np condition))
	position 
	new-discourse-referent)
    (when np
      (setq new-discourse-referent (gensym)
	    position (position np condition :test #'equal))
      (push new-discourse-referent (universe d))
      (push (list (second (first (last np))) new-discourse-referent)
	    (simple-conditions d))
      (setf (nth position condition) new-discourse-referent)
      (add-feature d  new-discourse-referent (first (last np)))
      condition)))


;; (d) Analyse von indefinite Descriptions
(defun id-np (struktur)
  (let ((np (my-assoc 'np struktur)))
    (if (member (second (second np))
		'(ein eine einer eines einen einem))
	np)))

(defmethod apply-id-rule ((d drs) condition)
  (let ((np (id-np condition))
	position 
	new-discourse-referent)
    (when np
      (setq new-discourse-referent (gensym)
	    position (position np condition :test #'equal))
      (push new-discourse-referent (universe d))
      (setf (nth position condition) new-discourse-referent)
      (if (listp (second (first (last np)))) ;; id mit rc
	  (push (cons (list 'nrc  new-discourse-referent) (rest (third np)))
		(complex-conditions d))
	  (push (list (second (first (last np))) new-discourse-referent)
		(simple-conditions d)))
      (add-feature d new-discourse-referent (second (third np)))
      condition)))

;; (e) Analyse von Relativstzen
(defmethod apply-rc-rule ((d drs) condition)
   (when (my-assoc 'nrc condition)
    (push (list (second (second condition)) (second (first condition)))
	  (simple-conditions d))
    (add-feature d (second (first condition)) (second condition))
    (subst (second (first condition)) '(np nil) (third (third condition))
	   :test #'equal)))

;; (f) Analyse von pronominalen NPs
(defun pro-np (struktur)
  (let ((np (my-assoc 'np struktur)))
    (if (eq (first (second np)) 'pro)
	np)))

(defmethod apply-pro-rule ((d drs) condition)
  (let* ((np (pro-np condition))
	 (antezedenz (antecedenz d (my-assoc 'gen (features (second np)))))
	 position 
	 new-discourse-referent)
    (when (and np antezedenz)
      (setq new-discourse-referent (gensym)
	    position (position np condition :test #'equal))
      (push new-discourse-referent (universe d))
      (push (list antezedenz '= new-discourse-referent) (simple-conditions d))
      (setf (nth position condition) new-discourse-referent)
      (add-feature d  new-discourse-referent (first (last np)))
      condition)))

(defmethod antecedenz ((d drs) gen-feat)
  (let ((bereich (if (super d)
		     (append (simple-conditions (super d))
			     (simple-conditions d))
		     (simple-conditions d))))
  (dolist (x bereich)
     (if (and (not (drs-p x))
	      (eq (first x) 'gen)
	      (eq (fourth x) (second gen-feat)))
	 (return (second x))))))

;; (g) Behandlung nicht-analysierbarer Bedingungen
(defmethod apply-error-rule ((d drs) condition)
  (format t "~%Die Bedingung ~a" condition)
  (format t "~%ist mit den vorhandenen Regeln nicht reduzierbar!")
  (list '!!!_irreduzierbare_drs_bedingung!!!))



;; Kleinere Funktionen:
(defun features (eintrag)
  (rest (my-assoc (first eintrag)
		 (rest (assoc (second eintrag) *lexikon*)))))

(defmethod add-feature ((d drs) dr struktur)
  (dolist (av-pair (features struktur))
     (case (first av-pair)
       (gen (push (list 'gen  dr '= (second av-pair))
		  (simple-conditions d)))
       (t nil))))

		

;;; (F) Daten
   
(setf *syntax* (make-instance 'syntax
			      :vn '(s np vp1 vp rc)
			      :vt '(det n v aux pro pn rpro neg)
			      :r '((s np vp) (vp vp1) (vp aux neg vp1)
				   (vp1 v) (vp1 v np) (NP)
				   (np det n) (np pn) (np pro)
				   (n n rc) (rc rpro s)))
      (tabelle *syntax*) (erreichbarkeitstabelle *syntax*))

(setq *lexikon*
      '((der (det (gen mas)) (rpro (gen mas)))
	(die (det (gen fem)) (rpro  (gen fem)))
	(ein (det (gen mas)) (det (gen neu)))
	(einen (det (gen mas)))
	(eine (det (gen fem)))
	(er (pro (gen mas)))
	(ihn (pro (gen mas)))
	(sie (pro (gen fem)))
	(mann (n (gen mas)))
	(frau (n (gen fem)))
	(hans (pn (gen mas)))
	(bill (pn (gen mas)))
	(maria (pn (gen fem)))
	(hat (aux))
	(nicht (neg))
	(schlummert (v))
	(geschlummert (v))
	(sucht (v))
	(liebt (v))
	(gesucht (v))))

(defun my-assoc (a b)
   (find a b :key #'(lambda (x) (if (listp x) (first x)x))))