;;;;---------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter error-correcting Recognizer
;;;;
;;;;---------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : ahopet72.lsp
;;;; Berechnung der Chart durch Vorwaertsverkettung
;;;; Repraesentation der Chart durch eine Liste

;;;; Letzte Aenderung: 4.3.93

;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax011 :lexikon lexikon02)


;;;; NEUE FUNKTIONEN:
;;;;
;;;; ERRORS (Kanten Selektor)
;;;; TERMINAL-ERROR-PRODUCTION-P
;;;; SAME-EDGE-P
;;;; COMPUTE-ERRORS

;;;; Geaenderte Funktionen:
;;;;
;;;; NEUE-KANTE-P
;;;; BILDE-KANTE
;;;; BUILD



;;;;                        PROGRAMM
;;;;                        --------


(defun recognize_ear (satz &key (symbol (startsymbol)))
" Earley-basierter Recognizer fuer einfache kontextfreie Syntaxen."
  (dolist (kante (SETQ *CHART* (GENERIERE-CHART SATZ SYMBOL))
		(format t "~%Kein Parse gefunden !"))
    (when (and (passive-kante-p kante)
	       (eq (kopf kante) symbol)
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
          (return t))))


(defun generiere-chart (satz symbol &optional chart)
" Hauptfunktion zum Aufbau der Chart."
  (declare (special chart))
  (initialisiere-chart symbol)
  (dotimes (i (length satz) chart)
    (dolist (kante (lexikalische-kanten i (nth i satz)))
      (build kante))))


;;; Initialisierung der Chart

(defun initialisiere-chart (symbol)
" Generiert die Kantes fuer das Feld Chart(0, 0)."
  (dolist (regel (regeln))
    (when (eq (first regel) symbol)
          (build `((0 0) ,(first regel) nil ,(rest regel) 0)))))

(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen Kanten; d.h. eine
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list (list start (+ start 1)) x (list wort) nil 0))
	  (eintrag-kategorien wort)))


;;; Berechnung neuer Kanten

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt eine Kante als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Ist sie aktiv, wird EXPAND aufgerufen; sonst COMPLETE."
  (declare (special chart))
  (when (neue-kante-p kante)
    (if (aktive-kante-p kante)
        (expand kante)
        (complete kante))))


(defun expand (kante)
" EXPAND nimmt eine aktive Kante als Argument und generiert fuer jede Regel der
  Syntax mit der das naechste Symbol der aktiven Seite des Kantes expandiert
  werden kann, eine neue Kante."
    (dolist (x (expansionen (first (offen kante))))
      (build (list (list (ende kante) (ende kante)) (first (offen kante)) nil x 0))))


(defun complete (c-kante)
" COMPLETE nimmt als Argument eine passive Kante und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieser Kantes mit einer
  aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (dolist (kante chart)
    (when (matching-kanten-p kante c-kante)
          (build (bilde-kante kante c-kante)))))



;;;; Selektoren, Konstruktoren & Praedikate
;;;; --------------------------------------

;;; Kantenselektoren

(defun start (kante)
  (caar kante))

(defun ende (kante)
  (cadar kante))

(defun kopf (kante)
  (second kante))

(defun geschlossen (kante)
  (third kante))

(defun offen (kante)
  (fourth kante))

(defun errors (kante)
  (fifth kante))

;;; Praedikate


(defun aktive-kante-p (kante)
" Das Praedikat testet, ob kante ein aktives kante ist."
  (offen kante))

(defun passive-kante-p (kante)
" Das Praedikat testet, ob kante ein passives kante ist."
  (not (offen kante)))


(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob kante1 mit kante2 zu einem neuen kante verbunden
  werden kann."
  (and (aktive-kante-p kante1)
       (eql (ende kante1) (start kante2))
       (eq  (first (offen kante1)) (kopf kante2))))


(defun terminal-error-production-p (regel)
" Das Praedikat pruft, ob es sich bei REGEL um eine Fehler-Regel
  handelt oder nicht. "
  (or (string-equal (symbol-name (first regel)) "I")           ; Insertion-Regel
      (and (string-equal (symbol-name (first regel)) "E" :start1 0 :end1 1)
	   (or (endp (rest regel))                             ; Deletion-Regel
	       (string-not-equal (symbol-name (first regel))   ; Substitution-Regel
				 (symbol-name (second regel)) :start1 1)))))

(defun same-edge-p (k1 k2)
" Das Pradikat prueft ob beide Kante bis auf die Fehlerzahl
  identisch sind. "
  (equal (butlast k1) (butlast k2)))


(defun neue-kante-p (kante)
" Das Praedikat testet, ob die Kante schon in der Chart enthalten ist
  oder nicht und traegt sie zw. ihre Fehlerzahl gegebenenfalls in die
  Chart ein."
  (declare (special chart))
  (let ((old (first (member kante chart :test #'same-edge-p))))
    (cond ((not old) (push kante chart))
	  ((> (errors old) (errors kante))
	   (setf (fifth old) (errors kante)))
	  (t nil))))



;;; Kantenkonstruktor

(defun bilde-kante (kante1 kante2)
" Diese Funktion bildet aus den beiden Kanten kante1 und kante2 eine neue
  Kante - kante1 wird durch kante2 ergaenzt."
  (list (list (start kante1) (ende kante2))
	(kopf kante1)
	(append (geschlossen kante1) (list (kopf kante2)))
	(rest (offen kante1))
	(COMPUTE-ERRORS KANTE1 KANTE2)))

(defun compute-errors (kante1 kante2)
" Die Funktion berechnet die Fehlerzahl fuer die aus KANTE1 und KANTE2 neu
  gebildete Kante. "
  (if (terminal-error-production-p (cons (kopf kante2) (geschlossen kante2)))
      (+ 1 (errors kante1) (errors kante2))
      (+ (errors kante1) (errors kante2))))

