;;;;---------------------------------------------------------------------------
;;;
;;;                      Ein Chart-Recognizer
;;;                             nach
;;;                      Cocke-Kasami-Younger
;;;
;;;;---------------------------------------------------------------------------


;;;; File : cky-1.lsp

;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax7 :lexikon lexikon7)


;;; Hauptfunktion
;;; -------------

(defun recognize_cky (satz &optional (symbol (startsymbol)))
" Top-Level Funktion des Chart-Recognizers nach Cocke, Kasami I&
  Younger."
  (let ((chart (make-chart satz)))
    (when (member symbol (first chart)) t)))

;; Zur Generierung eines Leftparses aus der einfachen Analysematrix muss
;; nur das "T" in der WHEN-Form ersetzt werden durch den Funktionsaufruf:
;; (left-parse 0 (length satz) symbol satz chart)



;;; Generierung einer einfachen Analysematrix
;;; -----------------------------------------

(defun make-chart (satz)
" Generiert eine Chart fuer eine Syntax SYNTAX in Chomsky-Normalform unter
  Verwendung des Algorithmuses von Cocke, Kasami & Younger.
  Die Felder werden durch Listen repraesentiert, die mit einer Index-Liste
  beginnen, auf die ein/mehrere Eintraege folgen.
  Ein Eintrag besteht aus einer lexikalischen/syntaktischen Kategorie.
  Beispiel: ((1 3) NP N1 S)  repraesentiert das Feld mit den Koordinaten
            1 3 der Chart. Es enthaelt drei Eintraege."
  (let ((chart (lexical-items satz))
	(n (length satz)))	
    (when (> n 1)
          (compute-other-items n chart))))

(defun lexical-items (Satz)
" Generiert die 'unterste Zeile' der Chart; d.h die Felder(i, i+1),
  mit i = 0, 1, ..., Laenge(SATZ)-1.
  Fuer jede Wortform des Satzes werden alle passenden lexikalischen
  Kategorien in das entsprechende Feld eingetragen."
  (let ((items nil)
          (index 0))     
    (dolist (x satz items)
      (setf items (append items (list `((,index ,(1+ index))
                         ,@(eintrag-kategorien x))))
          index (1+ index)))))


(defun compute-other-items (n chart)
" Generiert die uebrigen Felder der Chart; d.h. die Felder(i, j),
  mit j - i > 1. Die Eintraege in diesen Feldern bestehen aus syntaktischen
  Kategorien der Syntax."
  (do ((new-items chart)
       (k 2 (1+ k)))
      ((> k n) new-items)
      (do* ((i 0 (1+ i))
	    (j (+ i k) (1+ j))
            (symbols (search-chart i j new-items)
		     (search-chart i j new-items)))
           ((> i (- n k)))
           (when symbols	
           	 (setf new-items (cons `((,i ,j) ,@symbols) new-items))))))

       
(defun search-chart (i j chart)
" Berechnet die Eintraege fuer das Feld(i, j) der Chart.
  Dazu muessen fuer k, i < k < j, alle (schon berechneten) Paare von Feldern
  Feld(i k), Feld(k, j) durchsucht werden."
  (let ((resultat nil))
    (dotimes (k (- j i 1) resultat)
       (setf resultat
             (append resultat	
		     (match (assoc (list i (+ k i 1)) chart :test #'equal)
			    (assoc (list (+ k i 1) j) chart :test #'equal)
			    (regeln)))))))

(defun match (feld-1 feld-2 regeln)
" Berechnet fuer zwei Listen von Kategorien die Liste der linken Seiten der
  Regeln, deren rechte Seite aus einer Kategorie aus FELD-1 und einer
  Kategorie aus FELD-2 besteht."
  (let ((resultat nil)
	(rules nil))
    (dolist (x (rest feld-1) resultat)
      (dolist (y (rest feld-2))
	(when (setf rules
                    (remove-if #'(lambda (z) (not (equal (rest z) (list x y))))
			       regeln))
              (setf resultat (append resultat (mapcar #'first rules))))))))




;;; Generierung der Strukturbeschreibungen
;;; --------------------------------------

;; Aus dem Recognizer erhaelt man einen Parser, der fuer jeden akzeptierten
;; Satz einen Linksparse generiert, indem in der Funktion RECOGNIZE_CKY die
;; letzte Zeile "(when (member symbol (first chart)) t)))" ersetzt wird durch:
;; "(pprint (left-parse 0 (length satz) symbol satz (regeln) chart))".

(defun left-parse (i j kat satz chart)
" Diese Funktion generiert aus einer einfachen Analysematrix einen Leftparse
  fuer SATZ."
  (if (= i (1- j))
      (print (nth (- j 1) satz))
      (let ((seiten (expansionen kat)))
     	(dolist (x seiten)
       	  (do ((k (1+ i) (+ k 1)))
              ((eq k j))
              (when (and (member (first x)
				 (assoc `(,i ,k) chart :test #'equal))
                 	 (member (second x)
				 (assoc `(,k ,j) chart :test #'equal)))
          	    (print (index (cons kat x) (regeln)))
          	    (left-parse i k (first x) satz chart)
          	    (left-parse k j (second x) satz chart)))))))

(defun index (regel regel-set)
" Diese Funktion berechnet den Index von REGEL; d.h. sie bestimmt, an wie-
  vielter Position REGEL in REGEL-SET vorkommt."
  (+ (- (length regel-set) (length (member regel regel-set :test #'equal)))
     1))
  
