;;;----------------------------------------------------------------------------
;;;
;;;                        Ein Chart-Parser
;;;                             nach
;;;                      Cocke-Kasami-Younger
;;;
;;;----------------------------------------------------------------------------

;;; File : cky-2.lsp

;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax7 :lexikon lexikon7)


;;; Hauptfunktion
;;; -------------

(defun parse_cky (satz &optional (symbol (startsymbol)))
" Top-Level Funktion des Chart-Parsers nach Cocke, Kasami & Younger."
  (let ((chart (make-chart satz)))
    (pprint chart)
    (terpri)   
    (if (assoc symbol (first chart))
     	(pprint (p-struktur symbol 0 (length satz) satz chart))
      		;; wahlweise (p-struktur symbol 0 (length satz) satz chart))
      		;; zur Berechnung gepackter Strukturen
     	(format t "Der Satz ist nicht wohlgeformt.~%"))))


;;; Berechnung der komplexen Analysematrix
;;; --------------------------------------

(defun make-chart (satz)
" Generiert eine Chart fuer eine Syntax SYNTAX in Chomsky-Normalform unter
  Verwendung des Algorithmuses von Cocke, Kasami & Younger.
  Die Felder werden durch Listen repraesentiert, die mit einer Index-Liste
  beginnen, auf die ein/mehrere Eintraege folgen.
  Ein Eintrag besteht aus einer lexikalischen/syntaktischen Kategorie.

  Beispiel: ((1 3) NP N1 S)  repraesentiert das Feld mit den Koordinaten
            1 3 der Chart. Es enthaelt drei Eintraege."
  (let ((chart (lexical-items satz))
     	(n (length satz)))
    (when (> n 1)
       	  (compute-other-items n chart))))

(defun lexical-items (satz)
" Generiert die 'unterste Zeile' der Chart; d.h die Felder(i, i+1),
  mit i = 0, 1, ... Laenge(SATZ)-1.
  Fuer jede Wortform des Satzes werden alle passenden lexikalischen
  Kategorien in das entsprechende Feld eingetragen."
  (let ((items nil)
     	(index 0))
    (dolist (x satz items)
      (setf items (append items (list `((,index ,(1+ index))
                         		,@(eintrag-kategorien x))))
	    index (1+ index)))))


(defun compute-other-items (n chart)
" Generiert die uebrigen Felder der Chart; d.h. die Felder(i, j),
  mit j - i > 1. Die Eintraege in diesen Feldern bestehen aus Listen
  der Form:
                  (X (Y1 i1 j1) (Y2 i2 j2))
  X, Y1 & Y2 sind synt. Kategorien der Syntax und i1, j1 bzw. i2 & j2
  Koordinaten von den Feldern, in denen es einen Eintrag fuer Y1 bzw. Y2 gibt."
  (do ((new-items chart)
       (k 2 (1+ k)))
      ((> k n) new-items)
      (do* ((i 0 (1+ i))
            (j (+ i k) (1+ j))
            (symbols (search-chart i j new-items)
               	     (search-chart i j new-items)))
           ((> i (- n k)))
           (when symbols
                 (setf new-items (cons `((,i ,j) ,@symbols) new-items))))))

       
(defun search-chart (i j chart)
" Berechnet die Eintraege fuer das Feld(i, j) der Chart.
  Dazu muessen fuer k , i < k < j, alle (schon berechneten) Paare von Feldern
  Feld(i k), Feld(k, j) durchsucht werden."
  (let ((resultat nil))
    (dotimes (k (- j i 1) resultat)
       (setf resultat
	     (append resultat
               	     (match (assoc (list i (+ k i 1)) chart :test #'equal)
                   	    (assoc (list (+ k i 1) j) chart :test #'equal)
                   	    (regeln)))))))

(defun match (feld-1 feld-2 regeln)
" Berechnet fuer zwei Listen von Kategorien die Liste der linken Seiten der
  Regeln, deren rechte Seite aus einer Kategorie aus KATEGORIEN-1 und einer
  Kategorie aus KATEGORIEN-2 besteht."
  (let ((resultat nil)
     	(rules nil))
    (dolist (x (get-categories feld-1) resultat)
      (dolist (y (get-categories feld-2))
     	(setf rules (remove-if #'(lambda (z) (not (equal (rest z) (list x y))))
                      	       regeln))
     	(when rules
              (setf resultat
              	    (append resultat
                   	    (mapcar #'(lambda (z)
                         		(list (first z)
                               		      (cons x (first feld-1))
                               		      (cons y (first feld-2))))
                        	    rules))))))))


(defun get-categories (feld)
" Extrahiert alle Kategorien, fuer die es einen Eintrag in FELD gibt."
  (if (symbolp (second feld))
      (rest feld)
      (make_set (mapcar #'first (rest feld)))))
           
(defun make_set (set &key (test #'eq))
  (cond ((endp set) nil)
	((member (first set) (rest set) :test test)
      	 (make_set (rest set)))
     	(t (cons (first set) (make_set (rest set))))))



;;; Generierung der Strukturbeschreibungen
;;; --------------------------------------

; A.

(defun struktur (kat i j satz chart)
" Diese Funktion generiert eine einfache Strukturbeschreibung aus einer kom-
  plexen Analysematrix."
  (if (= i (1- j))
      (list kat (nth (- j 1) satz))
      (let ((eintrag (assoc kat (assoc (list i j) chart :test #'equal))))
     	(list kat
              (struktur (first (second eintrag))
		        (second (second eintrag))
               		(third (second eintrag))
               		satz
               		chart)
              (struktur (first (third eintrag))
			(second (third eintrag))
               		(third (third eintrag))
               		satz
               		chart)))))


; B.

(defun p-struktur (symbol i j satz chart)
" Generiert eine Strukturbeschreibung aus CHART, in der lokale Ambiguitaeten
  'gepacked' repraesentiert werden."
  (if (= i (1- j))
      (list symbol (nth (1- j) satz))
      (let ((resultat nil))
     	(dolist (x (remove-if-not #'(lambda (y) (eq (first y) symbol))
                      		  (assoc (list i j) chart :test #'equal)))
	  (setf resultat     
		(cons (list (apply #'p-struktur `(,@(second x) ,satz ,chart))
                            (apply #'p-struktur `(,@(third x) ,satz ,chart)))
                      resultat)))
	(if (endp (rest resultat))
            (cons symbol (first resultat))
            (cons symbol resultat)))))
