;;;;--------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Parser
;;;;
;;;;--------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : earl-p-4.lsp
;;;; Strukturgenerierung ueber Pointer
;;;; local ambiguity packing
;;;; effiziente Behandlung von Tilgungsregeln
;;;; keinen Look-ahead (FIRST-RELATION)

;;;; Lade (wenn noetig) die zur Strukturgenerierung notwendigen 
;;;; Funktionen:
(unless (fboundp 'generiere-struktur)
  (lade-datei "pointer.lsp"))





;;;;                        PROGRAMM
;;;;                        --------

(defun parse_ear (satz &key (symbol (startsymbol)))
" Earley-basierter Parser fuer einfache kontextfreie Syntaxen.
  Strukturgenerierung durch Pointer.
  KANTE        : (dotted-rule koordinaten)
  DOTTED-RULE  : ((start ende) kopf geschlossen offen)
  KOORDINATEN  : ((p11 (p12 ...)) ....(pn1 (pn2 ...))), mit n >= 0
                 pi1 verweist auf die aktive und pi2 auf die passive Kante,
                 aus der KANTE gebildet wurde. "
  (let ((chart (generiere-chart satz symbol)))
    (auswerten chart symbol satz)))

(defun generiere-chart (satz symbol &optional chart)
" Hauptfunktion zum Aufbau der Chart."
  (declare (special chart))
  (initialisiere-chart symbol)
  (dotimes (i (length satz) chart)
    (dolist (kante (lexikalische-kanten i (nth i satz)))
      (build kante))))


;;; Initialisierung der Chart

(defun initialisiere-chart (symbol)
" Generiert die Kanten fuer das Feld Chart(0, 0)."
  (dolist (regel (regeln))
    (when (eq (first regel) symbol)
          (build `(((0 0) ,(first regel) nil ,(rest regel)) nil)))))

(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen Kanten; d.h. eine
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list (list (list start (+ start 1)) x nil nil)
		    (list wort)))
	  (eintrag-kategorien wort)))


;;; Berechnung neuer Kanten

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt eine KANTE als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Ist sie aktiv, wird EXPAND aufgerufen; sonst COMPLETE."
  (declare (special chart))
  (when (update-chart kante)
    (cond ((passive-kante-p kante) (complete kante))
	  (t (expand kante)
	     (when (member (first (offen kante)) *deleteable*)
	       (berechne-tilgung kante))))))


(defun expand (kante)
" EXPAND nimmt eine aktive KANTE als Argument und generiert fuer jede Regel 
  der Syntax mit der das naechste Symbol der aktiven Seite der KANTE expan-
  diert werden kann, eine neue KANTE."
    (dolist (x (expansionen (first (nth 3 (kern kante)))))
      (build (list (list (list (ende kante) (ende kante))
			 (first (offen kante))
			 nil
			 x)
		   nil))))

(defun complete (pkante)
" COMPLETE nimmt als Argument eine passive KANTE und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieser KANTE mit einer
  aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (dolist (akante chart)
    (when (matching-kanten-p akante pkante)
          (build (list (list (list (start akante) (ende pkante))
			     (kopf akante)
			     (append (geschlossen akante)
				     (list (kopf pkante)))
			     (rest (offen akante)))
		       (list (list (index akante chart)
				   (list (index pkante chart)))))))))

(defun berechne-tilgung (kante)
" Das erste Symbol des offenen Abschnitts von KANTE ist ein tilgbares Symbol.
  Es wird eine entsprechende passive Kante (E-KANTE) und eine (aktive/passive)
  Kante erzeugt, die durch Uebertrag dieses Symbols in den geschlossenen Ab-
  schnitt der Kante entsteht."
  (declare (special chart))
  (let ((e-kante `(((,(ende kante) ,(ende kante))
		    ,(first (offen kante)) nil nil) nil)))
    (when (neue-kante-p e-kante chart) (update-chart e-kante))
    (build (list (list (caar kante) 
		       (kopf kante)
		       (append (geschlossen kante)(list (first (offen kante))))
		       (rest (offen kante)))
		 (list (list (index kante chart)
			     (list (index e-kante chart))))))))


(defun update-chart (kante)
" Traegt KANTE in die Chart ein, sofern sie noch nicht in ihr enthalten ist;
  sonst werden die Pointer von KANTE in den alten Eintrag mit aufgenommen.
  UPDATE-CHART evaluiert in diesem Fall zu NIL (vgl. BUILD)."
  (declare (special chart))
 (when (neue-kante-p kante chart)
   (let ((alte-kante (assoc (kern kante) chart :test #'equal)))
     (if alte-kante
	 (not (setf (second alte-kante)
		    (add (first (pointer kante)) (pointer alte-kante))))
         (setf chart (append chart (list kante)))))))



;;; Selektoren, Konstruktoren & Praedikate

;;;---------------------------------------

;;; Selektoren

(defun start (kante)
  (caaar kante))

(defun ende (kante)
  (cadaar kante))

(defun kopf (kante)
  (second (kern  kante)))

(defun geschlossen (kante)
  (third (kern kante)))

(defun offen (kante)
  (fourth (kern kante)))

(defun kern (kante)
  (first kante))

(defun pointer (kante)
  (second kante))


;;; Praedikate

(defun aktive-kante-p (kante)
" Das Praedikat testet, ob KANTE ein aktives Kante ist."
  (offen kante))

(defun passive-kante-p (kante)
" Das Praedikat testet, ob KANTE ein passives Kante ist."
  (not (offen kante)))

(defun neue-kante-p (kante chart)
" Das Praedikat testet, ob der Kante schon in der Chart enthalten ist
  oder nicht."
  (if (or (endp (pointer kante))              ; zyklische Kante
	  (atom (first (pointer kante))))     ; lexikalische Kante
      (not (member kante chart :test #'equal))
      (not (member (first (pointer kante))
		   (second (assoc (kern kante) chart :test #'equal))
		   :test #'equal))))

(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob KANTE1 mit KANTE2 zu einem neuen KANTE verbunden
  werden kann."
  (and (aktive-kante-p kante1)
       (eql (ende kante1) (start kante2))
       (eq  (first (offen kante1)) (kopf kante2))))

(defun index (kante chart)
" Berechnet die Position an der KANTE in CHART vorkommt.
  Mehrfachvorkommen sind aufgrund des Konstruktionsverfahrens fuer CHART
  ausgeschlossen.
  Index des ersten KANTEs der Chart: 0."
  (let ((found (assoc (kern kante) chart :test #'equal)))
    (if found
	(- (length chart) (length (member found chart :test #'equal)))
	(break "INDEX : KANTE not found in Chart"))))






