;;;;---------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Recognizer
;;;;
;;;;---------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : earl-r-1.lsp
;;;; Berechnung der Chart durch Vorwaertsverkettung
;;;; Repraesentation der Chart durch eine Liste

;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax21 :lexikon lexikon2)



;;;;                        PROGRAMM
;;;;                        --------


(defun recognize_ear (satz &key (symbol (startsymbol)))
" Earley-basierter Recognizer fuer einfache kontextfreie Syntaxen."
  (dolist (kante (generiere-chart satz symbol) 
		(format t "~%Kein Parse gefunden !"))
    (when (and (passive-kante-p kante)
	       (eq (kopf kante) symbol)
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
          (return t))))


(defun generiere-chart (satz symbol &optional chart)
" Hauptfunktion zum Aufbau der Chart."
  (declare (special chart))
  (initialisiere-chart symbol)
  (dotimes (i (length satz) chart)
    (dolist (kante (lexikalische-kanten i (nth i satz)))
      (build kante)))
  (PPRINT CHART))


;;; Initialisierung der Chart

(defun initialisiere-chart (symbol)
" Generiert die Kantes fuer das Feld Chart(0, 0)."
  (dolist (regel (regeln))
    (when (eq (first regel) symbol)
          (build `((0 0) ,(first regel) nil ,(rest regel))))))

(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen Kanten; d.h. eine
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list (list start (+ start 1)) x (list wort) nil))
	  (eintrag-kategorien wort)))


;;; Berechnung neuer Kanten

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt eine Kante als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Ist sie aktiv, wird EXPAND aufgerufen; sonst COMPLETE."
  (declare (special chart))
  (when (neue-kante-p kante chart)
    (push kante chart)
    (if (aktive-kante-p kante)
        (expand kante)
        (complete kante))))


(defun expand (kante)
" EXPAND nimmt eine aktive Kante als Argument und generiert fuer jede Regel der
  Syntax mit der das naechste Symbol der aktiven Seite des Kantes expandiert
  werden kann, eine neue Kante."
    (dolist (x (expansionen (first (offen kante))))
      (build (list (list (ende kante) (ende kante)) (first (offen kante)) nil x))))


(defun complete (c-kante)
" COMPLETE nimmt als Argument eine passive Kante und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieser Kantes mit einer
  aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (dolist (kante chart)
    (when (matching-kanten-p kante c-kante)
          (build (bilde-kante kante c-kante)))))



;;;; Selektoren, Konstruktoren & Praedikate
;;;; --------------------------------------

;;; Kantenselektoren

(defun start (kante)
  (caar kante))

(defun ende (kante)
  (cadar kante))

(defun kopf (kante)
  (second kante))

(defun geschlossen (kante)
  (third kante))

(defun offen (kante)
  (fourth kante))


;;; Praedikate

(defun neue-kante-p (kante chart)
" Das Praedikat testet, ob der kante schon in der Chart enthalten ist
  oder nicht."
  (not (member kante chart :test #'equal)))

(defun aktive-kante-p (kante)
" Das Praedikat testet, ob kante ein aktives kante ist."
  (offen kante))

(defun passive-kante-p (kante)
" Das Praedikat testet, ob kante ein passives kante ist."
  (not (offen kante)))


(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob kante1 mit kante2 zu einem neuen kante verbunden
  werden kann."
  (and (aktive-kante-p kante1)
       (eql (ende kante1) (start kante2))
       (eq  (first (offen kante1)) (kopf kante2))))


;;; Kantenkonstruktor

(defun bilde-kante (kante1 kante2)
" Diese Funktion bildet aus den beiden Kanten kante1 und kante2 eine neue
  Kante - kante1 wird durch kante2 ergaenzt."
  (list (list (start kante1) (ende kante2))
	(kopf kante1)
	(append (geschlossen kante1) (list (kopf kante2)))
	(rest (offen kante1))))