;;;;---------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Recognizer
;;;;
;;;;---------------------------------------------------------------------------

;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : earl-r-2.lsp
;;;; Berechnung der Chart durch Vorwaertsverkettung
;;;; Verwaltung der Chart ueber einen zweidimensionalen Array

;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax21 :lexikon lexikon2)



;;;;                        PROGRAMM
;;;;                        --------

(defun recognize_ear (satz &key	(symbol (startsymbol)))
  " Earley-basierter Recognizer fuer einfache kontextfreie Syntaxen."
  (let* ((dimension (1+ (length satz)))
         (chart (make-array (list dimension dimension))))
    (declare (special chart))
    (dolist (kante (aref (generiere-chart satz symbol) 0 (1- dimension))
		  (format t "~%Kein Parse gefunden !"))
      (when (and (passive-kante-p kante)
		 (eq (first kante) symbol))
		 (return t)))))


(defun generiere-chart (satz symbol)
" Hauptfunktion zum Aufbau der Chart."
  (declare (special chart))
  (initialisiere-chart symbol)
  (dotimes (i (length satz) chart)
    (dolist (kante (lexikalische-kanten i (nth i satz)))
      (build kante))))


;;; Initialisierung der Chart

(defun initialisiere-chart (symbol)
" Generiert die Kanten fuer das Feld Chart(0, 0)."
  (dolist (regel (regeln))
    (when (eq (first regel) symbol)
          (build `((0 0) ,(first regel) nil ,(rest regel))))))

(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen Kanten; d.h. eine
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list (list start (+ start 1)) x (list wort) nil))
	  (eintrag-kategorien wort)))


;;; Berechnung neuer Kanten

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt eine Kante als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Ist sie aktiv, wird EXPAND aufgerufen; sonst COMPLETE."
  (declare (special chart))
  (when (neue-kante-p kante chart)
        (setf (aref chart (start kante) (ende kante))
              (cons (rest kante) (aref chart (start kante) (ende kante))))
    (if (aktive-kante-p kante)
        (expand kante)
        (complete kante))))


(defun expand (kante)
" EXPAND nimmt eine aktive Kante als Argument und generiert fuer jede Regel der
  Syntax mit der das naechste Symbol der aktiven Seite der Kante expandiert
  werden kann, eine neue Kante."
    (dolist (x (expansionen (first (offen kante))))
      (build (list (list (ende kante) (ende kante)) (first (offen kante)) nil x))))


(defun complete (pkante)
" COMPLETE nimmt als Argument eine passive Kante und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieser Kantes mit einer
  aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (dotimes (x (1+ (start pkante)))
    (dolist (akante (aref chart x (start pkante)))
      (when (matching-kanten-p akante (rest pkante))
        (build (list (list x (ende pkante))
                     (first akante)
                     (append (second akante) (list (second pkante)))
                     (rest (third akante))))))))



;;;; Selektoren, Konstruktoren & Praedikate
;;;; --------------------------------------

;;; Kantenselektoren

(defun start (kante)
  (caar kante))

(defun ende (kante)
  (cadar kante))

(defun kopf (kante)
  (second kante))

(defun geschlossen (kante)
  (third kante))

(defun offen (kante)
  (fourth kante))


;;; Praedikate

(defun neue-kante-p (kante chart)
" Das Praedikat testet, ob der Kante schon in der Chart enthalten ist
  oder nicht."
  (not (member (rest kante) (aref chart (caar kante) (cadar kante))
               :test #'equal)))

(defun aktive-kante-p (kante)
" Das Praedikat testet, ob kante ein aktives kante ist."
  (offen kante))

(defun passive-kante-p (kante)
" Das Praedikat testet, ob kante ein passives kante ist."
  (not (third kante)))

(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob KANTE1 mit KANTE2 zu einem neuen Kante verbunden werden
  kann."
  (and (third kante1)
       (eq (first (third kante1)) (first kante2))))

