;;;;--------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Parser
;;;;                      nach Winograd[83]
;;;;			 fuer ID/LP-Syntaxen
;;;;
;;;;--------------------------------------------------------------------------

;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : idlp-p-2.lsp
;;;; Strukturgenerierung ueber Pointer
;;;; local ambiguity packing
;;;; effiziente Behandlung von Tilgungsregeln
;;;; keinen Look-ahead (FIRST-RELATION)

;;;; Lade (wenn noetig) die zur Verarbeitung von ID/LP-Syntaxen und zur
;;;; Strukturgenerierung notwendigen Funktionen:
(unless (fboundp 'id-regeln)
  (lade-datei "id-tools.lsp"))


(unless (fboundp 'generiere-struktur)
  (lade-datei "pointer.lsp"))


;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax8 :lexikon lexikon8)



;;;;                        PROGRAMM
;;;;                        --------


(defun parse_idlp (satz &key (symbol (startsymbol)))
" Earley-basierter Parser fuer einfache kontextfreie Syntaxen.
  Strukturgenerierung durch Pointer.
  KANTE        : (dotted-rule koordinaten)
  DOTTED-RULE  : (start ende kopf geschlossen offen)
  KOORDINATEN  : ((p11 (p12 ...)) ....(pn1 (pn2 ...))), mit n >= 0
                 pi1 verweist auf die aktive und pi2 auf die passivee Kante,
                 aus der kante gebildet wurde. "
  (do* ((chart (initialisiere-chart satz))
	(p-kanten (initialisiere_p-kanten symbol) (rest p-kanten))
	(kante (first p-kanten) (first p-kanten)))
       ((endp p-kanten) (auswerten chart symbol satz))
    (when (neue-kante-p kante chart)
      (setf chart (update-chart chart kante)
	    p-kanten (neue-kanten kante p-kanten chart)))))




;;; Initialisierung von CHART & KANTEN

(defun initialisiere-chart (satz)
" Diese Funktion traegt alle lexikalische Kanten fuer die Worte des Satzes in
  die Chart ein."
  (let ((resultat nil) (index 0))
    (dolist (wort satz resultat)
      (setf resultat 
	    (append resultat (lexikalische-kanten index wort))
	    index (1+ index)))))

(defun lexikalische-kanten (start wort)
" generiert fuer ein wort alle moeglichen lexikalischen kanten; d.h. einen
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list (list start (+ start 1) x nil nil) (list wort)))
	  (eintrag-kategorien wort)))

(defun initialisiere_p-kanten (symbol)
" Diese Funktion initailisiert die Chart, indem sie alle S-Kanten eintraegt."
  (let ((resultat nil))
    (dolist (regel (id-regeln) resultat)
      (when (eq (first regel) symbol)
	    (push (list (list 0 0 (first regel) nil (rest regel)) ())
		  resultat)))))



;;; Aktualisierung von CHART & KANTEN

(defun update-chart (chart kante)
" Diese Funktion traegt KANTE in CHART ein."
  (let ((alte-kante (assoc (kern kante) chart :test #'equal)))
    (cond (alte-kante
	   (setf (pointer alte-kante)
		 (add (first (pointer kante)) (pointer alte-kante)))
	   chart)
	  (t (append chart (list kante))))))

(defun add (verweis verweisliste)
" Diese Funktion ergaenzt die Liste der Pointer-Paare einer Kante mit einem
  weiteren Pointer-Paar."
  (let ((x (assoc (first verweis) verweisliste))) 
    (if x
	(if (member (first (second verweis)) (second x))
	    verweisliste
	    (cons (list (first verweis) (append (second verweis) (second x)))
		   (remove x verweisliste :test #'equal)))
	(cons verweis verweisliste))))

(defun neue-kanten (kante p-kanten chart)
" Diese Funktion kombiniert P-kanten mit den durch EXPAND und COMPLETE gene-
  rierten neuen Kanten. Erweiterungen dieser Funktion erlaubt die Realisierung
  verschiedener Suchstrategien."
  (let ((alte-kanten (append p-kanten chart)))
    (append p-kanten
	    (remove-if #'(lambda (x) (member x alte-kanten :test #'equal))
		       (append (complete chart kante)
			       (when (aktive-kante-p kante)
				 (expand kante)))))))



;;; Berechnung neuer Kanten


(defun expand (kante)
" EXPAND nimmt ein aktives kante als Argument und generiert fuer jede Regel 
  der Syntax mit der das naechste Symbol der aktiven Seite des kanten expan-
  diert werden kann, ein neues kante."
  (mapcar #'(lambda (x)
              (list (list (ende kante) (ende kante) (first x) nil (rest x))
                    nil))
	  (zulaessige-regeln (offen kante))))


(defun complete (chart kante)
" COMPLETE nimmt als Argument eine (aktive/passivee) Kante und traegt in die
  Chart diejenigen Kanten ein, die sich durch Verbindung dieser Kante mit
  einer (passiven/aktiven) Kante der Chart bilden lassen." 
  (if (passive-kante-p kante)
      (mapcar #'(lambda (x) (bilde-kante x kante chart))
	      (remove-if #'(lambda (x) (or (passive-kante-p x)
					   (not (matching-kanten-p x kante))))
			 chart))
      (mapcar #'(lambda (x) (bilde-kante kante x chart))
	      (remove-if #'(lambda (x) (or (aktive-kante-p x)
					   (not (matching-kanten-p kante x))))
			 chart))))



;;; Selektoren, Konstruktoren & Praedikate

;;;---------------------------------------

;;; Selektoren

(defun start (kante)
  (caar kante))

(defun ende (kante)
  (cadar kante))

(defun kopf (kante)
  (third (kern  kante)))

(defun geschlossen (kante)
  (fourth (kern kante)))

(defun offen (kante)
  (fifth (kern kante)))

(defun kern (kante)
  (first kante))

(defmacro pointer (kante)
  `(second ,kante))

;;; Praedikate

(defun aktive-kante-p (kante)
" Das Praedikat testet, ob kante ein aktives kante ist."
  (offen kante))

(defun passive-kante-p (kante)
" Das Praedikat testet, ob kante ein passivees kante ist."
  (endp (offen kante)))

(defun neue-kante-p (kante chart)
  (if (or (endp (pointer kante))		; zyklische Kanten
	  (atom (first (pointer kante))))	; lexikalische Kanten
      (not (assoc (kern kante) chart :test #'equal))
      (not (member (first (second (first (pointer kante))))
		   (second (assoc (caar (pointer kante))
	    (pointer (assoc (kern kante) chart :test #'equal))))))))
 


(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob kante1 mit kante2 zu einem neuen kante verbunden
  werden kann."
  (and (eql (ende kante1) (start kante2))
       (member (kopf kante2) (zulaessige-symbole (offen kante1)))))


;;; Kantenkonstruktor

; Geaendert:
(defun bilde-kante (kante1 kante2 chart)
" Diese Funktion bildet aus den beiden Kanten kante1 und kante2 eine neue
  Kante - kante1 wird durch kante2 ergaenzt."
  (list (list (start kante1)
	      (ende kante2)
	      (kopf kante1)
	      (append (geschlossen kante1) (list (kopf kante2)))
	      (entferne (kopf kante2) (offen kante1)))
	(list (list (index kante1 chart) (list (index kante2 chart))))))

(defun index (kante chart)
" Berechnet die Position an der kante in CHART vorkommt.
  Mehrfachvorkommen sind aufgrund des Konstruktionsverfahrens fuer CHART
  ausgeschlossen.
  Index des ersten kanten der Chart: 0."
  (let ((found (assoc (kern kante) chart :test #'equal)))
    (if found
	(- (length chart) (length (member found chart :test #'equal)))
	(break "INDEX : kante not found in Chart"))))


;-----------------------------------------------------------------------------

#|    Notwendige Aenderungen, um mit Bartons Repraesentation zu parsen:
      -----------------------------------------------------------------

; Geaenderte Funktionen:

(defun neue-kante-p (kante chart)
  (not (alte-kante-p kante chart)))

(defun update-kante (kante chart)
" Diese Funktion ergaenzt die Verweisliste der Kante in CHART, deren Kern
  mit dem von KANTE identisch ist."
  (let ((alte-kante (alte-kante-p kante chart)))		; hier
    (setf (pointer alte-kante)
	  (add (first (pointer kante)) (pointer alte-kante)))))

; Neue Funktionen:

(defun alte-kante-p (kante chart)
" ALTE-KANTE-P prueft,ob es in CHART eine zu KANTE aequivalente Kante gibt.
  Funktionswert: diese Kante / NIL. "
  (or (assoc (kern kante) chart :test #'equal)
      (dolist (x chart)
	(when (and (eq (start kante) (start x))
		   (eq (ende kante) (ende x))
		   (eq (kopf kante) (kopf x))
		   (eq-sets-p (geschlossen kante) (geschlossen x)))
	  (return x)))))


(defun eq-sets-p (l1 l2)
" Dieses Praedikat prueft, ob L1 und L2 identische Mengen darstellen. " 
  (when (eq (length l1) (length l2))
    (dolist (x l1 t)
      (unless (member x l2)
	(return nil)))))

|#
