;;;;---------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Recognizer
;;;;			 	   fuer
;;;;		              ID/LP-Syntaxen
;;;;
;;;;---------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : idlp-r-1.lsp

;;;; Lade (wenn noetig) die zur Verarbeitung von ID/LP-Syntaxen:
(unless (fboundp 'id-regeln)
  (lade-datei "id-tools.lsp"))


;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax8 :lexikon lexikon8)



;;;;                        PROGRAMM
;;;;                        --------


(defun recognize_idlp (satz &key (symbol (startsymbol)))
" Earley-basierter Recognizer fuer einfache kontextfreie Syntaxen."
  (dolist (kante (generiere-chart satz symbol) 
		(format t "~%Kein Parse gefunden !"))
    (when (and (passive-kanten-p kante)
	       (eq (kopf kante) symbol)
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
          (return t))))


(defun generiere-chart (satz symbol &optional chart)
" Hauptfunktion zum Aufbau der Chart."
  (declare (special chart))
  (initialisiere-chart symbol)
  (dotimes (i (length satz) chart)
    (dolist (kante (lexikalische-kanten i (nth i satz)))
      (build kante))))


;;; Initialisierung der Chart

(defun initialisiere-chart (symbol)
" Generiert die Kanten fuer das Feld Chart(0, 0)."
  (dolist (regel (regeln))
    (when (eq (first regel) symbol)
          (build `((0 0) ,(first regel) nil ,(rest regel))))))

(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen Kanten; d.h. eine
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list (list start (+ start 1)) x (list wort) nil))
	  (eintrag-kategorien wort)))


;;; Berchnung der Chart

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt eine Kante als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Ist sie aktiv, wird EXPAND aufgerufen; sonst COMPLETE."
  (declare (special chart))
  (when (neue-kante-p kante chart)
    (push kante chart)
    (if (aktive-kanten-p kante)
        (expand kante)
        (complete kante))))

; Geaendert:
(defun expand (kante)
" EXPAND nimmt eine aktive Kante als Argument und generiert fuer jede Regel der
  Syntax mit der das naechste Symbol der aktiven Seite der Kanten expandiert
  werden kann, eine neue Kante. "
  (dolist (x (zulaessige-regeln (offen kante)))      ; hier
    (build (list (list (ende kante) (ende kante)) 
		 (first x) nil (rest x)))))


;;; Geaendert:
(defun complete (pkante)
" COMPLETE nimmt als Argument ein passives Kante und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieses Kanten mit einem
  aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (dolist (akante chart)
    (when (matching-kanten-p akante pkante)
          (build (bilde-kante akante pkante)))))



;;;; Selektoren, Konstruktoren & Praedikate
;;;; --------------------------------------

;;; Kantenselektoren

(defun start (kante)
  (caar kante))

(defun ende (kante)
  (cadar kante))

(defun kopf (kante)
  (second kante))

(defun geschlossen (kante)
  (third kante))

(defun offen (kante)
  (fourth kante))


;;; Praedikate

;; Geaendert:
(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob KANTE1 mit KANTE2 zu einer neuen Kante verbunden werden
  kann."
  (and (aktive-kanten-p kante1)
       (eql (ende kante1) (start kante2))
       (member (kopf kante2) (zulaessige-symbole (offen kante1)))))

(defun neue-kante-p (kante chart)
" Das Praedikat testet, ob die Kante schon in der Chart enthalten ist
  oder nicht."
  (not (member kante chart :test #'equal)))

(defun aktive-kanten-p (kante)
" Das Praedikat testet, ob KANTE eine aktive Kante ist."
  (offen kante))

(defun passive-kanten-p (kante)
" Das Praedikat testet, ob KANTE eine passive Kante ist."
  (not (offen kante)))

(defun bilde-kante (akante pkante)
  (list (list (start akante) (ende pkante))
	(kopf akante)
	(append (geschlossen akante) (list (kopf pkante)))
	(entferne (kopf pkante) (offen akante))))