;;;;--------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Recognizer
;;;;			 fuer ID/LP-Syntaxen
;;;;                      nach Winograd[83]
;;;;
;;;;--------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : idlp-r-2.lsp
;;;; keinen Look-ahead (FIRST-RELATION)

;;;; Lade (wenn noetig) die zur Verarbeitung von ID/LP-Syntaxen:
(unless (fboundp 'id-regeln)
  (lade-datei "id-tools.lsp"))


;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax8 :lexikon lexikon8)



;;;;                        PROGRAMM
;;;;                        --------


(defun recognize_idlp (satz &key (symbol (startsymbol)))
" Earley-basierter Recognizer fuer einfache kontextfreie Syntaxen.
  kante        : (start ende kopf geschlossen offen) "
  (do* ((chart (initialisiere-chart satz))
	(p-kanten (initialisiere_p-kanten symbol))
	(kante (first p-kanten) (first p-kanten)))
       ((endp p-kanten) (auswerten chart symbol satz))
    (setf chart (update-chart chart kante)
	  p-kanten (kombiniere (rest p-kanten)
                            chart
			    (complete chart kante)
			    (when (aktive-kante-p kante)
			          (expand kante))))))


;;; Aktualisierung von CHART und KANTEN:

(defun update-chart (chart kante)
  (append chart (list kante)))

(defun kombiniere (p-kanten chart neue-kanten1 neue-kanten2)
" Diese Funktion kombiniert P-kanten mit den durch EXPAND und COMPLETE gene-
  rierten neuen Kanten. Erweiterungen dieser Funktion erlaubt die Realisierung
  verschiedener Suchstrategien."
  (let ((alte-kanten (append p-kanten chart)))
    (append p-kanten
            (remove-if #'(lambda (x) (member x alte-kanten :test #'equal))
                       (append neue-kanten1 neue-kanten2)))))


;;; Initialisierung von CHART und KANTEN

(defun initialisiere-chart (satz)
" Diese Funktion traegt alle lexikalische Kanten fuer die Worte des Satzes in
  die Chart ein."
  (let ((resultat nil) (index 0))
    (dolist (wort satz resultat)
      (setf resultat (append resultat (lexikalische-kanten index wort))
	    index (1+ index)))))

(defun lexikalische-kanten (start wort)
" generiert fuer ein wort alle moeglichen lexikalischen kanten; d.h. einen
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list start (+ start 1) x nil nil))
	  (eintrag-kategorien wort)))

(defun initialisiere_p-kanten (symbol)
" Diese Funktion initailisiert die Chart, indem sie alle S-Kanten eintraegt."
  (let ((resultat nil))
    (dolist (regel (id-regeln) resultat)
      (when (eq (first regel) symbol)
	    (push (list 0 0 (first regel) nil (rest regel)) resultat)))))


;;; Berechnung des Ergebnisses

(defun auswerten (chart symbol satz)
" Diese Funktion generiert alle Strukturbeschreibungen fuer alle Kanten vom
  Typ SYMBOL, die sich ueber den ganzen Satz erstrecken."
  (dolist (kante chart)
    (when (and (passive-kante-p kante)
	       (eq (kopf kante) symbol)
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
          (return t))))



;;; Berechnung neuer Kanten


(defun expand (kante)
" EXPAND nimmt ein aktives kante als Argument und generiert fuer jede Regel 
  der Syntax mit der das naechste Symbol der aktiven Seite des kanten expan-
  diert werden kann, ein neues kante."
  (mapcar #'(lambda (x)
              (list (ende kante) (ende kante) (first x) nil (rest x)))
	  (zulaessige-regeln (offen kante))))


(defun complete (chart kante)
" COMPLETE nimmt als Argument eine (aktive/passive) Kante und traegt in die
  Chart diejenigen Kanten ein, die sich durch Verbindung dieser Kante mit
  einer (passiven/aktiven) Kante der Chart bilden lassen." 
  (if (passive-kante-p kante)
      (mapcar #'(lambda (x) (bilde-kante x kante))
	      (remove-if #'(lambda (x)
			     (or (passive-kante-p x)
				 (not (matching-kanten-p x kante))))
			 chart))
      (mapcar #'(lambda (x) (bilde-kante kante x))
	      (remove-if #'(lambda (x)
			     (or (aktive-kante-p x)
				 (not (matching-kanten-p kante x))))
			 chart))))



;;;; Selektoren, Konstruktoren & Praedikate
;;;; --------------------------------------

;;; Kantenselektoren

(defun start (kante)
  (first kante))

(defun ende (kante)
  (second kante))

(defun kopf (kante)
  (third kante))

(defun geschlossen (kante)
  (fourth kante))

(defun offen (kante)
  (fifth kante))


;;; Praedikate

(defun neue-kante-p (kante chart)
" Das Praedikat testet, ob der kante schon in der Chart enthalten ist
  oder nicht."
  (not (member kante chart :test #'equal)))

(defun aktive-kante-p (kante)
" Das Praedikat testet, ob kante ein aktives kante ist."
  (offen kante))

(defun passive-kante-p (kante)
" Das Praedikat testet, ob kante ein passives kante ist."
  (not (offen kante)))

; Geaendert:
(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob kante1 mit kante2 zu einem neuen kante verbunden
  werden kann."
  (and (eql (ende kante1) (start kante2))
       (member (kopf kante2) (zulaessige-symbole (offen kante1))))) ; hier


;;; Kantenkonstruktor

; Geaendert:
(defun bilde-kante (kante1 kante2)
" Diese Funktion bildet aus den beiden Kanten kante1 und kante2 eine neue
  Kante - kante1 wird durch kante2 ergaenzt."
  (list (start kante1)
	(ende kante2)
	(kopf kante1)
	(append (geschlossen kante1) (list (kopf kante2)))
	(entferne (kopf kante2) (offen kante1))))  ; hier



