;;;;--------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Insel-Parser
;;;;
;;;;--------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : inselp-2.lsp
;;;; Strukturgenerierung ueber Pointer
;;;; local ambiguity packing
;;;; keine effiziente Behandlung von Tilgungs- & Kettenregeln
;;;; keinen Look-ahead (FIRST-RELATION)

;;;; Lade (wenn noetig) die zum Insel-Parsen notwendigen Hilfsfunktionen:
(unless (fboundp 'inseln)
  (lade-datei "is-tools.lsp"))


;;;; Lade (wenn noetig) die zur Strukturgenerierung notwendigen 
;;;; Funktionen:
(unless (fboundp 'generiere-struktur)
  (lade-datei "pointer.lsp"))


;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax21 :lexikon lexikon2)



;;;;                        PROGRAMM
;;;;                        --------


(defun parse_il (satz &key (symbol (startsymbol)))
" Earley-basierter Insel-Recognizer fuer einfache kontextfreie Syntaxen.
	KANTE := (start ende kopf links geschlossen rechts struktur) "
  (let* ((kanten (inseln satz)) (chart (second kanten)))
    (declare (special chart))
    (dolist (kante (first kanten) chart) 
      (build kante))  
    (auswerten chart symbol satz)))


;;; Initialisierung der Chart

(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen Kanten; d.h. einen
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list start (+ start 1) x nil (list wort) nil (list wort)))
	  (eintrag-kategorien wort)))


;;; Berechnung der Chart

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt ein Item als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Generiert COMPLETE keine neuen Kanten, wird EXPAND (KANTE ist aktiv) oder
  REDUCE (KANTE ist passiv) aufgerufen. "
  (declare (special chart))
  (when (update-chart kante)
	(unless (complete kante)
	  (if (aktive-kante-p kante)
	      (expand kante)
	      (reduce kante)))))


(defun expand (kante)
" EXPAND nimmt eine aktive Kante als Argument und generiert fuer jede Regel
  der Syntax mit der das naechste Symbol der aktiven Seite der Kante expan-
  diert werden kann, eine neue Kante."
  (if (rechts kante)
      (dolist (x (expansionen (first (rechts kante))))
        (build (list (ende kante) (ende kante)
		     (first (rechts kante)) nil nil x nil)))
      (dolist (x (expansionen (first (last (links kante)))))
	(build (list (start kante) (start kante)
		     (first (last (links kante))) x nil nil nil)))))


(defun complete (kante)
" COMPLETE nimmt als Argument eine aktive/passive Kante und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieser Kante mit einer
  passiven/aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (let ((wert nil))
    (if (aktive-kante-p kante)
	(dolist (k chart wert)
	  (when (and (passive-kante-p k) (matching-kanten-p kante k))
	    (unless wert (setq wert t))
	    (build (bilde-kante kante k))))
        (dolist (k chart wert)
	  (when (and (aktive-kante-p k) (matching-kanten-p k kante))
	    (unless wert (setq wert t))
	    (build (bilde-kante k kante)))))))
	

(defun reduce (kante)
" REDUCE nimmt als Argument eine passive Kante und generiert fuer jedes
  Vorkommen des Kopfs dieser Kante auf der rechten Seite einer der Regeln der 
  Syntax eine neue (ausser bei unaeren Regeln) aktive Kante. "
  (let ((kat (kopf kante)))
    (dolist (x (regeln))
      (dotimes (y (length (rest x)))
	(when (eq (nth y (rest x)) kat)
	  (build (list (start kante) 
		       (ende kante)
		       (first x)
		       (butlast (rest x) (- (length (rest x)) y))
		       (list kat)
		       (nthcdr (+ y 2) x)
		       (list (list -1 (list (index kante chart)))))))))))


;;;; Kantenkonstruktor & Pointerselektor

(defun bilde-kante (kante1 kante2)
" Durch Verbindung der aktiven Kante KANTE1 mit der passiven Kante KANTE2 wird
  eine neue aktive/passive Kante erzeugt. "
  (if (rechts kante1)
      (list (start kante1) 
	    (ende kante2)
	    (kopf kante1)
	    (links kante1)
	    (append (geschlossen kante1) (list (kopf kante2)))
	    (rest (rechts kante1))
	    (list (list (index kante1 chart)
			(LIST (index kante2 chart)))))
      (list (start kante2) 
	    (ende kante1)
	    (kopf kante1)
	    (butlast (links kante1))
	    (append (list (kopf kante2)) (geschlossen kante1))
	    (rechts kante1)
	    (list (list (index kante1 chart)
			(LIST (index kante2 chart)))))))



(defun pointer (kante)
  (seventh kante))


;;;; Aktualisierung der Chart

(defun update-chart (kante)
" Traegt KANTE in die Chart ein, sofern er noch nicht in ihr enthalten ist;
  sonst werden die Pointer von KANTE in den alten Eintrag mit aufgenommen."
  (declare (special chart))
  (let ((alte-kante 
	  (dolist (k chart)
	    (when (equal (butlast k) (butlast kante))
	      (return k)))))
    (if alte-kante
	(when (pointer kante)
	      (setf (seventh alte-kante)
	 	    (add (first (pointer kante)) (pointer alte-kante))))
	(setf chart (append chart (list kante))))))

(defun index (kante chart)
"  Berechnet die Position an der KANTE in CHART vorkommt.
  Mehrfachvorkommen sind aufgrund des Konstruktionsverfahrens fuer CHART
  ausgeschlossen.
  Index des ersten Items der Chart: 0."
  (let ((index 0))
    (dolist (k chart 
	       (cerror "Unkalkulierbare Ergebnisse."
		      "Die Kante ~a ist nicht in der Chart enthalten!" 
		      kante))
      (if (equal (butlast kante) (butlast k))
	  (return index)
	  (setf index (1+ index))))))


;-----------------------------------------------------------------------------

;;; Strukturgenerierung
;;; -------------------

(defun kombiniere-strukturen (kante structure1 structure2)
  (cond ((endp structure1) structure2)
        ((and (= (length (geschlossen kante)) (1+ (length structure1)))
	      (listp (first structure1)))
	 (if (match-p (first (geschlossen kante)) (first structure1))
	     (append structure1 (list structure2))
	     (append (list structure2) structure1)))
	(t (if (match-p (first (geschlossen kante)) (first structure1))
	       (list structure1 structure2)
 	       (list structure2 structure1)))))

(defun match-p (symbol struktur)
  (if (atom struktur)
      (eq symbol struktur)
      (match-p symbol (first struktur))))


