;;;;--------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Insel-Parser
;;;;                          nach Winograd[83]
;;;;
;;;;--------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;; File : inselp-3.lsp
;;;; Strukturgenerierung ueber Pointer
;;;; local ambiguity packing
;;;; effiziente Behandlung von Tilgungsregeln
;;;; keinen Look-ahead (FIRST-RELATION)

;;;; Lade (wenn noetig) die zum Insel-Parsen notwendigen Hilfsfunktionen:
(unless (fboundp 'inseln)
  (lade-datei "is-tools.lsp"))


;;;; Lade (wenn noetig) die zur Strukturgenerierung notwendigen 
;;;; Funktionen:
(unless (fboundp 'generiere-struktur)
  (lade-datei "pointer.lsp"))


;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax21 :lexikon lexikon2)



;;;;                        PROGRAMM
;;;;                        --------

(defun parse_il (satz &key (symbol (startsymbol)))
" Earley-basierter Parser fuer einfache kontextfreie Syntaxen.
  Strukturgenerierung durch Pointer.
  KANTE        : (start ende kopf links geschlossen rechts koordinaten)
  KOORDINATEN  : ((p11 (p12 ...)) ....(pn1 (pn2 ...))), mit n >= 0
                 pi1 verweist auf die aktive und pi2 auf die passivee Kante,
                 aus der kante gebildet wurde. "
  (do* ((l-kanten (inseln satz))
	(kante (caar l-kanten) (first p-kanten))
	(p-kanten (rest (first l-kanten)) (rest p-kanten))
	(chart (second l-kanten))
	(alte-kante (alte-kante-p kante chart)
		    (alte-kante-p kante chart)))
       ((endp kante) (auswerten chart symbol satz))
    (if alte-kante
	(update-kante alte-kante kante)
	(setf chart    (update-chart chart kante)
	      p-kanten (neue-kanten kante p-kanten chart)))))


;;; Initialisierung der Chart

(defun lexikalische-kanten (start wort)
" generiert fuer ein wort alle moeglichen lexikalischen kanten; d.h. einen
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list start (+ start 1) x nil (list wort) nil (list wort)))
	  (eintrag-kategorien wort)))



;;; Aktualisierung von CHART & KANTEN

(defun update-chart (chart kante)
" Diese Funktion traegt KANTE in CHART ein."
  (append chart (list kante)))

(defun update-kante (alte-kante kante)
" Diese Funktion ergaenzt die Verweisliste der Kante in CHART, deren Kern
  mit dem von KANTE identisch ist."
  (when (pointer kante)
    (setf (seventh alte-kante)
      (add (first (pointer kante)) (pointer alte-kante)))))

(defun neue-kanten (kante p-kanten chart)
" Diese Funktion kombiniert P-kanten mit den durch COMPLETE und EXPAND gene-
  rierten neuen Kanten. Erweiterungen dieser Funktion erlaubt die Realisierung
  verschiedener Suchstrategien."
  (let ((alte-kanten (append p-kanten chart)))
    (append (remove-if #'(lambda (x) (member x alte-kanten :test #'equal))
		       (or (complete chart kante)
			   (if (aktive-kante-p kante)
			       (expand kante)
			       (reduce kante chart))))
	    p-kanten)))



;;; Berechnung neuer Kanten

(defun expand (kante)
" EXPAND nimmt ein aktives kante als Argument und generiert fuer jede Regel 
  der Syntax mit der das naechste Symbol der aktiven Seite des kanten expan-
  diert werden kann, ein neues kante."
  (let ((resultat nil))
    (if (rechts kante)
	(dolist (x (expansionen (first (rechts kante))) resultat)
	  (push (list (ende kante) (ende kante)
		      (first (rechts kante)) nil nil x nil)
		resultat))
        (dolist (x (expansionen (first (last (links kante)))) resultat)
	  (push (list (start kante) (start kante)
		      (first (last (links kante))) x nil nil nil)
		resultat)))))



(defun complete (chart kante)
" COMPLETE nimmt als Argument eine (aktive/passivee) Kante und traegt in die
  Chart diejenigen Kanten ein, die sich durch Verbindung dieser Kante mit
  einer (passiven/aktiven) Kante der Chart bilden lassen." 
  (if (passive-kante-p kante)
      (mapcar #'(lambda (x) (bilde-kante x kante chart))
	      (remove-if #'(lambda (x) (or (passive-kante-p x)
					   (not (matching-kanten-p x kante))))
			 chart))
      (mapcar #'(lambda (x) (bilde-kante kante x chart))
	      (remove-if #'(lambda (x) (or (aktive-kante-p x)
					   (not (matching-kanten-p kante x))))
			 chart))))

(defun reduce (kante chart)
" REDUCE nimmt als Argument eine passive Kante und generiert fuer jedes
  Vorkommen des Kopfs dieser Kante auf der rechten Seite einer der Regeln der 
  Syntax eine neue (ausser bei unaeren Regeln) aktive Kante. "
  (let ((kat (kopf kante)) (resultat nil))
    (dolist (x (regeln) resultat)
      (dotimes (y (length (rest x)))
	(when (eq (nth y (rest x)) kat)
	  (push (list (start kante) 
		      (ende kante)
		      (first x)
		      (butlast (rest x) (- (length (rest x)) y))
		      (list kat)
		      (nthcdr (+ y 2) x)
		      (list (list -1 (list (index kante chart)))))
		 resultat))))))


(defun index (kante chart)
"  Berechnet die Position an der KANTE in CHART vorkommt.
  Mehrfachvorkommen sind aufgrund des Konstruktionsverfahrens fuer CHART
  ausgeschlossen.
  Index des ersten Items der Chart: 0."
  (let ((index 0))
    (dolist (k chart 
	       (cerror "Unkalkulierbare Ergebnisse."
		      "Die Kante ~a ist nicht in der Chart enthalten!" 
		      kante))
      (if (equal (butlast kante) (butlast k))
	  (return index)
	  (setf index (1+ index))))))



;;; Selektoren, Konstruktoren & Praedikate

;;;---------------------------------------

;;; Selektor

(defun pointer (kante)
  (seventh kante))


;;; Praedikat

(defun alte-kante-p (kante chart)
  (dolist (k chart)
    (when (equal (butlast k) (butlast kante))
       (return k))))


;;; Kantenkonstruktor

(defun bilde-kante (kante1 kante2 chart)
" Durch Verbindung der aktiven Kante KANTE1 mit der passiven Kante KANTE2 wird
  eine neue aktive/passive Kante erzeugt. "
  (if (rechts kante1)
      (list (start kante1) 
	    (ende kante2)
	    (kopf kante1)
	    (links kante1)
	    (append (geschlossen kante1) (list (kopf kante2)))
	    (rest (rechts kante1))
	    (list (list (index kante1 chart)
			(list (index kante2 chart)))))
      (list (start kante2) 
	    (ende kante1)
	    (kopf kante1)
	    (butlast (links kante1))
	    (append (list (kopf kante2)) (geschlossen kante1))
	    (rechts kante1)
	    (list (list (index kante1 chart)
			(list (index kante2 chart)))))))


;-----------------------------------------------------------------------------

;;; Strukturgenerierung
;;; -------------------

(defun kombiniere-strukturen (kante structure1 structure2)
  (cond ((endp structure1) structure2)
        ((and (= (length (geschlossen kante)) (1+ (length structure1)))
	      (listp (first structure1)))
	 (if (match-p (first (geschlossen kante)) (first structure1))
	     (append structure1 (list structure2))
	     (append (list structure2) structure1)))
	(t (if (match-p (first (geschlossen kante)) (first structure1))
	       (list structure1 structure2)
 	       (list structure2 structure1)))))

(defun match-p (symbol struktur)
  (if (atom struktur)
      (eq symbol struktur)
      (match-p symbol (first struktur))))


