;;;;--------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Insel-Parser
;;;;                          nach Winograd[83]
;;;;
;;;;--------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;; File : inselp-4.lsp

;;;;   Hilfsfunktionen
;;;;   ---------------


(defmacro matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob KANTE1 mit KANTE2 zu einer neuen KANTE verbunden
   werden kann."
  `(if (rechts ,kante1)
       (and (eql (ende ,kante1) (start ,kante2))
	    (eq (first (rechts ,kante1)) (kopf ,kante2)))
       (and (eql (start ,kante1) (ende ,kante2))
	    (eq (first (last (llinks ,kante1))) (kopf ,kante2)))))

(defmacro neue-kante-p (kante chart)
" Das Praedikat testet, ob der KANTE schon in der Chart enthalten ist
  oder nicht."
  `(not (member ,kante ,chart :test #'equal)))

(defmacro aktive-kante-p (kante)
" Das Praedikat testet, ob KANTE ein aktives KANTE ist."
  `(or (llinks ,kante) (rechts ,kante)))

(defmacro passive-kante-p (kante)
" Das Praedikat testet, ob KANTE ein passives KANTE ist."
  `(not (aktive-kante-p ,kante)))


(defmacro start (kante)
  `(first ,kante))

(defmacro ende (kante)
  `(second ,kante))

(defmacro kopf (kante)
  `(third ,kante))

(defmacro geschlossen (kante)
  `(fifth ,kante))

(defmacro llinks (kante)
  `(fourth ,kante))

(defmacro rechts (kante)
  `(sixth ,kante))

(defmacro struktur (kante)
  `(seventh ,kante))

(defmacro luecken (kante)
  `(eighth ,kante))



;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax21 :lexikon lexikon2)

; Funktionen zur Bestimmung der Inseln:

(defun inseln (satz)
" Diese Funktion erlaubt es dem Benutzer, wahlweise Worte des Satzes oder
  lexikalische Kategorien als Inseln fuer die Analyse zu markieren, bzw. mit
  der default-Inselkategorie (V) zu arbeiten. "
  (let ((antwort nil))
    (format t "~%Waehlen Sie eine der folgenden Optionen:~%")
    (format t "    (A)  -  dialoggesteuerte Festlegung von Wort-Inseln~%")
    (format t "    (B)  -  dialoggesteuerte Festlegung von Insel-Kategorien~%")
    (format t "    (C)  -  Verwendung der Kategorie V als Insel.~%> ")
    (setf antwort (read))
    (cond ((eq antwort 'a) (wort-inseln satz))
	  ((eq antwort 'b) (insel-kategorien satz))
	  (t (insel-kategorien satz '(V) )))))

(defun wort-inseln (satz)
" Diese Funktion legt durch einen mit dem Benutzer gefuehrten Dialog fest,
  welche Worte des Satzes als Insel verwendet werden."
  (let ((z 0) (heads nil) (noheads nil))
    (dolist (x satz (list heads noheads))
      (format t "~%Soll / ~A / als Insel verwendet werden?   " (nth z satz))
      (if (y-or-n-p)
	  (setf heads (append heads (lexikalische-kanten z x)))
	  (setf noheads (append noheads (lexikalische-kanten z x))))
      (setf z (1+ z)))))

(defun insel-kategorien (satz &optional default)
" Diese Funktion legt durch einen mit dem Benutzer gefuehrten Dialog fest,
  welche lexikalische Kategorien als Insel verwendet werden."
  (let ((z 0) (kategorien nil) (heads nil) (noheads nil))
    (unless (setf kategorien default)
      (format t "~%Welche der folgenden lexikalischen Kategorien:  ~{~S ~}~%" 
	        (lexikalische-kategorien))
      (format t "sollen als Insel verwendet werden? ~%>  ")
      (setf kategorien (read)) (terpri))
    (dolist (x satz (list heads noheads))
      (dolist (y (lexikalische-kanten z x))
	(if (member (kopf y) kategorien) 
	    (push y heads)
	    (push y noheads)))
      (setf z (1+ z)))))



;;;;                        PROGRAMM
;;;;                        --------

(defun parse (satz &key (symbol (startsymbol)))
" Earley-basierter Parser fuer einfache kontextfreie Syntaxen.
  KANTE        : (start ende kopf llinks geschlossen rechts struktur) "
  (do* ((l-kanten (inseln satz))
	(kante (caar l-kanten) (first p-kanten))
	(p-kanten (rest (first l-kanten)) (rest p-kanten))
	(chart (second l-kanten)))
       ((endp kante) (auswerten chart symbol satz))
	(setf chart    (update-chart chart kante)
	      p-kanten (neue-kanten kante p-kanten chart))))


;;; Initialisierung der Chart

(defun lexikalische-kanten (start wort)
" generiert fuer ein wort alle moeglichen lexikalischen kanten; d.h. einen
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list start (+ start 1) x nil (list wort) nil (list wort)))
	  (eintrag-kategorien wort)))



;;; Aktualisierung von CHART & KANTEN

(defun update-chart (chart kante)
" Diese Funktion traegt KANTE in CHART ein."
  (append chart (list kante)))

(defun neue-kanten (kante p-kanten chart)
" Diese Funktion kombiniert P-kanten mit den durch COMPLETE und EXPAND gene-
  rierten neuen Kanten. Erweiterungen dieser Funktion erlaubt die Realisierung
  verschiedener Suchstrategien."
  (let ((alte-kanten (append p-kanten chart)))
    (append (remove-if #'(lambda (x) (member x alte-kanten :test #'equal))
		       (or (complete chart kante)
			   (if (aktive-kante-p kante)
			       (expand kante)
			       (rreduce kante chart))))
	    p-kanten)))


;;; Auswerten der Chart

(defun auswerten (chart symbol satz)
" Diese Funktion generiert alle Strukturbeschreibungen fuer alle Kanten vom
  Typ SYMBOL, die sich ueber den ganzen Satz erstrecken."
  (dolist (kante chart)
    (when (and (passive-kante-p kante)
	       (eq (kopf kante) symbol)
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
      (pprint (cons (kopf kante) (struktur kante))))))


;;; Berechnung neuer Kanten

(defun expand (kante)
" EXPAND nimmt ein aktives kante als Argument und generiert fuer jede Regel 
  der Syntax mit der das naechste Symbol der aktiven Seite des kanten expan-
  diert werden kann, ein neues kante."
  (let ((resultat nil))
    (if (rechts kante)
	(dolist (x (expansionen (first (rechts kante))) resultat)
	  (push (list (ende kante) (ende kante)
		      (first (rechts kante)) nil nil x nil)
		resultat))
        (dolist (x (expansionen (first (last (llinks kante)))) resultat)
	  (push (list (start kante) (start kante)
		      (first (last (llinks kante))) x nil nil nil)
		resultat)))))



(defun complete (chart kante)
" COMPLETE nimmt als Argument eine (aktive/passivee) Kante und traegt in die
  Chart diejenigen Kanten ein, die sich durch Verbindung dieser Kante mit
  einer (passiven/aktiven) Kante der Chart bilden lassen." 
  (if (passive-kante-p kante)
      (mapcar #'(lambda (x) (bilde-kante x kante chart))
	      (remove-if #'(lambda (x) (or (passive-kante-p x)
					   (not (matching-kanten-p x kante))))
			 chart))
      (mapcar #'(lambda (x) (bilde-kante kante x chart))
	      (remove-if #'(lambda (x) (or (aktive-kante-p x)
					   (not (matching-kanten-p kante x))))
			 chart))))

(defun rreduce (kante chart)
" REDUCE nimmt als Argument eine passive Kante und generiert fuer jedes
  Vorkommen des Kopfs dieser Kante auf der rechten Seite einer der Regeln der 
  Syntax eine neue (ausser bei unaeren Regeln) aktive Kante. "
  (let ((kat (kopf kante)) (resultat nil))
    (dolist (x (regeln) resultat)
      (dotimes (y (length (rest x)))
	(when (eq (nth y (rest x)) kat)
	  (push (list (start kante) 
		      (ende kante)
		      (first x)
		      (butlast (rest x) (- (length (rest x)) y))
		      (list kat)
		      (nthcdr (+ y 2) x)
		      (list (CONS KAT (STRUKTUR KANTE))))
		 resultat))))))


;;; Kanten-Konstruktor
;;;-------------------

(defun bilde-kante (kante1 kante2 chart)
" Durch Verbindung der aktiven Kante KANTE1 mit der passiven Kante KANTE2 wird
  eine neue aktive/passive Kante erzeugt. "
  (if (rechts kante1)
      (list (start kante1) 
	    (ende kante2)
	    (kopf kante1)
	    (llinks kante1)
	    (append (geschlossen kante1) (list (kopf kante2)))
	    (rest (rechts kante1))
	    (APPEND (STRUKTUR KANTE1)
		    (LIST (CONS (KOPF KANTE2) (STRUKTUR KANTE2)))))
      (list (start kante2) 
	    (ende kante1)
	    (kopf kante1)
	    (butlast (llinks kante1))
	    (append (list (kopf kante2)) (geschlossen kante1))
	    (rechts kante1)
	    (APPEND (LIST (CONS (KOPF KANTE2) (STRUKTUR KANTE2)))
		    (STRUKTUR KANTE1)))))





