;;;;---------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Insel-Recognizer
;;;;			     mit Vorwaertsverkettung
;;;;---------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------


;;;; File : inselr-1.lsp


;;;; Lade (wenn noetig) die zum Insel-Parsen notwendigen Hilfsfunktionen:
(unless (fboundp 'inseln)
  (lade-datei "is-tools.lsp"))

;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax21 :lexikon lexikon2)



;;;;                        PROGRAMM
;;;;                        --------


(defun recognize_il (satz &key (symbol (startsymbol)))
" Earley-basierter Insel-Recognizer fuer einfache kontextfreie Syntaxen.
	KANTE := (start ende kopf links geschlossen rechts) "
  (let* ((kanten (inseln satz)) (chart (second kanten)))
    (declare (special chart))
    (dolist (kante (first kanten) chart) 
      (build kante))
    (pprint chart)
    (auswerten chart symbol satz)))


(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen KANTEs; d.h. einen
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list start (+ start 1) x nil (list wort) nil))
	  (eintrag-kategorien wort)))


;;; Auswertung der Chart:

(defun auswerten (chart symbol satz)
" Diese Funktion generiert alle Strukturbeschreibungen fuer alle Kanten vom
  Typ SYMBOL, die sich ueber den ganzen Satz erstrecken."
  (dolist (kante chart)
    (when (and (passive-kante-p kante)
	       (eq (kopf kante) symbol)
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
          (return t))))


;;; Berechnung neuer Kanten:

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt eine Kante als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Ist sie aktiv, wird EXPAND aufgerufen; sonst COMPLETE."
  (declare (special chart))
  (when (neue-kante-p kante chart)
    (push kante chart)
    (unless (complete kante)
      (if (aktive-kante-p kante)
	  (expand kante)
          (reduce kante)))))


(defun expand (kante)
" EXPAND nimmt eine aktive Kante als Argument und generiert fuer jede Regel
  der Syntax mit der das naechste Symbol der aktiven Seite der Kante expan-
  diert werden kann, eine neue Kante."
  (if (rechts kante)
      (dolist (x (expansionen (first (rechts kante))))
        (build (list (ende kante) (ende kante)
		     (first (rechts kante)) nil nil x)))
      (dolist (x (expansionen (first (last (links kante)))))
	(build (list (start kante) (start kante)
		     (first (last (links kante))) x nil nil)))))

(defun reduce (kante)
" REDUCE nimmt als Argument eine passive Kante und generiert fuer jedes
  Vorkommen des Kopfs dieser Kante auf der rechten Seite einer der Regeln der 
  Syntax eine neue (ausser bei unaeren Regeln) aktive Kante. "
  (let ((kat (kopf kante)))
    (dolist (x (regeln))
      (dotimes (y (length (rest x)))
	(when (eq (nth y (rest x)) kat)
	  (build (list (start kante) 
		       (ende kante)
		       (first x)
		       (butlast (rest x) (- (length (rest x)) y))
		       (list kat)
		       (nthcdr (+ y 2) x))))))))


(defun complete (kante)
" COMPLETE nimmt als Argument eine aktive/passive Kante und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieser Kante mit einer
  passiven/aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (let ((wert nil))
    (if (aktive-kante-p kante)
	(dolist (k chart wert)
	  (when (and (passive-kante-p k) (matching-kanten-p kante k))
	    (unless wert (setq wert t))
	    (build (bilde-kante kante k))))
        (dolist (k chart wert)
	  (when (and (aktive-kante-p k) (matching-kanten-p k kante))
	    (unless wert (setq wert t))
	    (build (bilde-kante k kante)))))))

	


;;; Konstruktor, Praedikate und Selektoren

(defun bilde-kante (kante1 kante2)
  (if (rechts kante1)
      (list (start kante1) 
	    (ende kante2)
	    (kopf kante1)
	    (links kante1)
	    (append (geschlossen kante1) (list (kopf kante2)))
	    (rest (rechts kante1)))
      (list (start kante2) 
	    (ende kante1)
	    (kopf kante1)
	    (butlast (links kante1))
	    (append (list (kopf kante2)) (geschlossen kante1))
	    (rechts kante1))))




