;;;;--------------------------------------------------------------------------
;;;;
;;;;             Ein earley-basierter Insel-Recognizer
;;;;                      nach Winograd[83]
;;;;
;;;;--------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : inselr-2.lsp
;;;; keinen Look-ahead (FIRST-RELATION)
;;;;
;;;; letzte Aenderung: 4.9.92  -  NEUE-KANTEN: (APPEND (complete ... ) (if ...) )
  
;;;; Lade (wenn noetig) die zum Insel-Parsen notwendigen Hilfsfunktionen:
(unless (fboundp 'inseln)
  (lade-datei "is-tools.lsp"))

;;;; Waehle eine geeignete Syntax und ein geeignetes Lexikon:
(use :syntax syntax21 :lexikon lexikon2)



;;;;                        PROGRAMM
;;;;                        --------


(defun recognize_il (satz &key (symbol (startsymbol)))
" Earley-basierter Insel-Recognizer fuer einfache kontextfreie Syntaxen.
         kante =: (start ende kopf links geschlossen rechts) "
  (do* ((l-kanten (inseln satz))
	(kante (caar l-kanten) (first p-kanten))
        (p-kanten (rest (first l-kanten)) (rest p-kanten))
	(chart (second l-kanten)))
       ((endp kante) (pprint chart) (auswerten chart symbol satz))
    (setf chart (update-chart chart kante)
	  p-kanten (neue-kanten kante p-kanten chart))))



;;; Aktualisierung und Initialisieren von CHART und KANTEN:

(defun update-chart (chart kante)
  (append chart (list kante)))

(defun neue-kanten (kante p-kanten chart)
" Diese Funktion kombiniert P-kanten mit den durch EXPAND und COMPLETE gene-
  rierten neuen Kanten. Erweiterungen dieser Funktion erlaubt die Realisierung
  verschiedener Suchstrategien."
  (let ((alte-kanten (append p-kanten chart)))
    (append p-kanten
            (remove-if #'(lambda (x) (member x alte-kanten :test #'equal))
		       (append (complete chart kante)
			   (if (aktive-kante-p kante)
			       (expand kante)
			       (reduce kante)))))))


(defun lexikalische-kanten (start wort)
" generiert fuer ein wort alle moeglichen lexikalischen kanten; d.h. einen
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list start (+ start 1) x nil (list wort) nil))
	  (eintrag-kategorien wort)))


;;; Auswerung von CHART

(defun auswerten (chart symbol satz)
" Diese Funktion generiert alle Strukturbeschreibungen fuer alle Kanten vom
  Typ SYMBOL, die sich ueber den ganzen Satz erstrecken."
  (dolist (kante chart)
    (when (and (passive-kante-p kante)
	       (eq (kopf kante) symbol)
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
          (return t))))



;;; Berechnung neuer Kanten

(defun expand (kante)
" EXPAND nimmt ein aktives kante als Argument. Ist die rechte Abschnitt der
  Kante nicht leer, wird fuer jede Regel der Syntax mit der das naechste Symbol
  dieses Abschnitts expandiert werden kann, eine neue Kante generiert; sonst
  fuer das letzte Symbol des rechten Abschnitts."
  (if (rechts kante)
      (mapcar #'(lambda (x) (list (ende kante) (ende kante) 
				  (first (rechts kante)) nil nil x))
	      (expansionen (first (rechts kante))))
      (mapcar #'(lambda (x) (list (start kante) (start kante) 
				  (first (last (links kante))) x nil nil))
	      (expansionen (first (last (links kante)))))))


(defun reduce (kante)
" REDUCE nimmt als Argument eine passive Kante und generiert fuer jedes
  Vorkommen des Kopfs dieser Kante auf der rechten Seite einer der Regeln der 
  Syntax eine neue (ausser bei unaeren Regeln) aktive Kante. "
  (let ((resultat nil) 
	(kat (kopf kante)))
    (dolist (x (regeln) resultat)
      (dotimes (y (length (rest x)))
	(when (eq (nth y (rest x)) kat)
	  (push (list (start kante) (ende kante) (first x)
		      (butlast (rest x) (- (length (rest x)) y))
		      (list kat)
		      (nthcdr (+ y 2) x))
		resultat))))))
	

(defun complete (chart kante)
" complete nimmt als Argument eine (aktive/passive) Kante und traegt in die
  Chart diejenigen Kanten ein, die sich durch Verbindung dieser Kante mit
  einer (passiven/aktiven) Kante der Chart bilden lassen." 
  (if (passive-kante-p kante)
      (mapcar #'(lambda (x) (bilde-kante x kante))
	      (remove-if #'(lambda (x)
			     (or (passive-kante-p x)
				 (not (matching-kanten-p x kante))))
			 chart))
      (mapcar #'(lambda (x) (bilde-kante kante x))
	      (remove-if #'(lambda (x)
			     (or (aktive-kante-p x)
				 (not (matching-kanten-p kante x))))
			 chart))))



;;;; Kantenkonstruktor
;;;; -----------------

(defun bilde-kante (kante1 kante2)
" Diese Funktion bildet aus den beiden Kanten kante1 und kante2 eine neue
  Kante: kante1 wird durch kante2 (rechts/links) ergaenzt."
  (if (rechts kante1)
      (list (start kante1)
	    (ende kante2)
	    (kopf kante1)
	    (links kante1)
	    (append (geschlossen kante1) (list (kopf kante2)))
	    (rest (rechts kante1)))
      (list (start kante2)
	    (ende kante1)
	    (kopf kante1)
	    (butlast (links kante1))
	    (append (list (kopf kante2)) (geschlossen kante1))
	    (rechts kante1))))



















