;;;;---------------------------------------------------------------------------
;;;;
;;;;                  Ein earley-basierter Recognizer
;;;;                             fuer
;;;;                   unifikationsbasierte Syntaxen
;;;;
;;;;---------------------------------------------------------------------------


;;;;                    INFOS & INITIALISIERUNG
;;;;                    -----------------------

;;;; File : pearl-r.lsp
;;;; letzte Aenderung: 29.5.92

;;;; Berechnung der Chart durch Vorwaertsverkettung
;;;; Repraesentation der Chart durch eine Liste

;;;; geaenderte Funktionen:
;;;;
;;;; RECOGNIZE_EAR
;;;; INITIALISIERE-CHART
;;;; LEXIKALISCHE-KANTEN
;;;; EXPAND
;;;; MATCHING-KANTEN-P
;;;; BILDE-KANTEN
;;;; NEUE-KANTE-P


;;;; neue Funktionen:
;;;; ----------------
;;;;
;;;; KOMPATIBLE-REGELN (vorher: EXPANSIONEN)
;;;; LINKE-SEITE (Makro)
;;;; RECHTE-SEITE (Makro)
;;;; VARIABLEN
;;;; REGEL-SUBSUMPTION
;;;; KOPIERE-KANTE



;;;;                        PROGRAMM
;;;;                        --------


(defun recognize_ear (satz &key (symbol (list (startsymbol) nil)))
" Earley-basierter Recognizer fuer einfache kontextfreie Syntaxen."
  (dolist (kante (generiere-chart satz symbol) 
		(format t "~%Kein Parse gefunden !"))
    (when (and (passive-kante-p kante)
	       (unify (list (kopf kante) (variablen kante)) symbol)  ; HIER
	       (eq (start kante) 0)
	       (eq (ende kante) (length satz)))
      (pprint kante)
      (pprint-dag (cons (kopf kante) (geschlossen kante)) (variablen kante))
      (return))))



(defun generiere-chart (satz symbol &optional chart)
" Hauptfunktion zum Aufbau der Chart."
  (declare (special chart))
  (initialisiere-chart symbol)
  (dotimes (i (length satz) (progn (pprint chart)chart))
    (dolist (kante (lexikalische-kanten i (nth i satz)))
      (build kante))))


;;; Initialisierung der Chart

(defun initialisiere-chart (symbol)
" Generiert die Kantes fuer das Feld Chart(0, 0)."
  (dolist (regel (regeln))
    (when (unify (list (linke-seite regel) (second regel)) symbol)     ; VON HIER AN
          (build `((0 0) ,(linke-seite regel) nil 
			 ,(rechte-seite regel) ,(second regel))))))

(defun lexikalische-kanten (start wort)
" Generiert fuer ein Wort alle moeglichen lexikalischen Kanten; d.h. eine
  pro zugeordneter lexikalischer Kategorie."
  (mapcar #'(lambda (x)
	      (list (list start (+ start 1)) (first x)  (list (list (list 'phon wort))) nil 
		    (second x)))  ; HIER
	  (eintrag-kategorien wort)))


;;; Berechnung neuer Kanten

(defun build (kante)
" 'Arbeitspferd' das fuer den Aufbau der Chart sorgt.
  Nimmt eine Kante als Argument.
  Wenn KANTE noch nicht in der Chart enthalten ist, wird sie eingetragen.
  Ist sie aktiv, wird EXPAND aufgerufen; sonst COMPLETE."
  (declare (special chart))
  (when (neue-kante-p kante chart)
    (let ((neue-kante (kopiere-kante kante)))
      (push neue-kante chart)
      (if (aktive-kante-p neue-kante)
	  (expand neue-kante)
          (complete neue-kante)))))


(defun expand (kante)
" EXPAND nimmt eine aktive Kante als Argument und generiert fuer jede Regel der
  Syntax mit der das naechste Symbol der aktiven Seite des Kantes expandiert
  werden kann, eine neue Kante."
    (dolist (x (kompatible-regeln
		 (list (first (offen kante)) (variablen kante)))) ; VON HIER AN
      (build (list (list (ende kante) (ende kante)) 
		   (linke-seite x) nil (rechte-seite x) (second x)))))


(defun kompatible-regeln (symbol)
; SYMBOL ist eine nicht-terminale Kategorie von SYNTAX
; als Wert wird die Liste aller Regeln geliefert, deren linke
; Seite mit SYMBOL unifiziert
  (let ((resultat nil) (unifikator nil))
    (dolist (x (regeln) resultat)
      (setq unifikator (unify (list (linke-seite x) (second x)) symbol))
      (when unifikator 
	(push (list (cons (first unifikator) (rechte-seite x))
		    (second unifikator))
	      resultat)))))


(defun complete (c-kante)
" COMPLETE nimmt als Argument eine passive Kante und traegt in die Chart
  diejenigen Kanten ein, die sich durch Verbindung dieser Kantes mit einer
  aktiven Kante der Chart bilden lassen." 
  (declare (special chart))
  (let ((match nil))
    (dolist (kante chart)
      (setq match (matching-kanten-p kante c-kante))
      (when match
	(build (bilde-kante kante c-kante match))))))



;;;; Selektoren, Konstruktoren & Praedikate
;;;; --------------------------------------

;;; Kantenselektoren

(defun start (kante)
  (caar kante))

(defun ende (kante)
  (cadar kante))

(defun kopf (kante)
  (second kante))

(defun geschlossen (kante)
  (third kante))

(defun offen (kante)
  (fourth kante))

(defun variablen (kante)
  (fifth kante))




(defmacro linke-seite (regel)
  `(caar ,regel))

(defmacro rechte-seite (regel)
  `(cdar ,regel))


;;; Praedikate

(defun aktive-kante-p (kante)
" Das Praedikat testet, ob kante ein aktives kante ist."
  (offen kante))

(defun passive-kante-p (kante)
" Das Praedikat testet, ob kante ein passives kante ist."
  (not (offen kante)))


(defun matching-kanten-p (kante1 kante2)
" Das Praedikat testet, ob kante1 mit kante2 zu einem neuen kante verbunden
  werden kann."
  (and (aktive-kante-p kante1)
       (eql (ende kante1) (start kante2))
       (unify  (list (first (offen kante1)) (variablen kante1)) ; HIER
	       (list (kopf kante2) (variablen kante2)))))


;;;; NEUE-KANTE-P wurde komplett reformuliert:

(defun neue-kante-p (kante chart)
" Das Praedikat testet, ob KANTE von einer anderen in der Chart enthaltenen
  Kante subsumiert wird oder nicht."
  (flet ((aehnliche-kante (x kante)
	   (and (= (start x) (start kante))
		(= (ende x) (ende kante))
		(= (length (offen x)) (length (offen kante)))
		(= (length (geschlossen x)) (length (geschlossen kante))))))
    (dolist (k chart t)
      (if (and (aehnliche-kante k kante)
	       (regel-subsumption (cons (kopf k) (append (offen k) (geschlossen k)))
				  (variablen k)
				  (cons (kopf kante) 
					(append (offen kante) (geschlossen kante)))
				  (variablen kante)))
	       (return nil)))))


;;;; REGEL-SUBSUMPTION
;;;; testet, ob REGEL1 REGEL2 subsumiert

(defun regel-subsumption (regel1 bindung1 regel2 bindung2)
  (every #'(lambda (x y)
	     (subsumes1 x bindung1 y bindung2))
	 regel1
	 regel2))



;;; Kantenkonstruktor

(defun bilde-kante (kante1 kante2 unifikator)     ; HIER
" Diese Funktion bildet aus den beiden Kanten kante1 und kante2 eine neue
  Kante - kante1 wird durch kante2 ergaenzt."
    (list (list (start kante1) (ende kante2))
	  (kopf kante1)
	  (append (geschlossen kante1)
		  (list (first unifikator)))      ; HIER
	  (rest (offen kante1))
	  (second unifikator)))


(defun kopiere-kante (kante)
  (let ((regel (kopiere-regel (cons (kopf kante) 
				    (append (geschlossen kante) (offen kante)))
			      (variablen kante))))
  (list (first kante) 
	(caar regel) 
	(butlast (rest (first regel)) (length (offen kante)))
	(nthcdr (length (geschlossen kante)) (rest (first regel)))
	(second regel))))
