;;;;--------------------------------------------------------
;;;;
;;;;               SUBSUMPTIONS-RELATION
;;;;
;;;;--------------------------------------------------------


;;;; Datei: subsumes.lsp
;;;; letzte Aenderung: 15.4.92
;;;;
;;;; Funktionen:
;;;; SUBSUMES/2
;;;; SUBSUME1/4
;;;; VALUE-SUBSUMES/4
;;;; VARIABLE-SUBSUMES/4
;;;; VAR-WERT/2
;;;; VAR-SUBSUME/4


;;;; --------------------------------------------------------------------------
;;;; Funktion       SUBSUMES
;;;;
;;;; Argument(e)  : zwei Dags: DAG1, DAG2
;;;; Wert(e)      : T/NIL
;;;; Beschreibung : Top-Level Funktion: prueft, ob DAG1 DAG2 subsumiert
;;;; --------------------------------------------------------------------------
(defun subsumes (dag1 dag2)
  (subsumes1 (pfade dag1) (bindungen dag1) (pfade dag2) (bindungen dag2)))


;;;; --------------------------------------------------------------------------
;;;; Funktion       SUBSUMES1
;;;;
;;;; Argument(e)  : zwei Pfadmengen DAG1, DAG2 und
;;;;                zwei Bindungslisten BINDUNG1, BINDUNG2
;;;; Wert(e)      : T/NIL
;;;; Beschreibung : Diese zentrale Funktion arbeitet alle Pfade in DAG1 ab:
;;;;                - gibt es einen Pfad in DAG1 zu dem es keinen korrespondie-
;;;;                  renden Pfad in DAG2 gibt, den er subsumiert, evaluiert
;;;;                  SUBSUMES1 zu NIL;
;;;;                - wenn alle Pfade in DAG1 abgearbeitet werden konnten zu T
;;;; --------------------------------------------------------------------------
(defun subsumes1 (dag1 bindung1 dag2 bindung2)
  (let* ((pfad1 (first dag1))
	 (pfad2 (assoc (first pfad1) dag2)))
  (cond ((endp dag1) t)
	(pfad2
	 (and (value-subsumes (wert pfad1) bindung1 (wert pfad2) bindung2)
	      (subsumes1 (rest dag1) bindung1 dag2 bindung2)))
	(t nil))))

;;;; --------------------------------------------------------------------------
;;;; Funktion       VALUE-SUBSUMES
;;;;
;;;; Argument(e)  : zwei Werte WERT1, WERT2 (Typ: Atom v Liste) und
;;;;                zwei Bindungslisten BINDUNG1, BINDUNG2
;;;; Wert(e)      : T gdw. WERT1 WERT2 subsumiert; sonst NIL
;;;; Beschreibung : Die Funktion ueberprueft, ob WERT1 den WERT2 subsumiert
;;;; --------------------------------------------------------------------------
(defun value-subsumes (wert1 bindung1 wert2 bindung2)
  (cond ((equal wert1 wert2) t)
	((or (variable-p wert1) (variable-p wert2))
	 (variable-subsumes wert1 bindung1 wert2 bindung2))
	((and (listp wert1) (listp wert2))
	 (subsumes1 wert1 bindung1 wert2 bindung2))
	(t nil)))
	
;;;; --------------------------------------------------------------------------
;;;; Funktion       VARIABLE-SUBSUMES
;;;;
;;;; Argument(e)  : zwei Werte WERT1, WERT2, von denen mindestens einer eine
;;;;                eine Variable ist
;;;; Wert(e)      : T gdw. WERT1 WERT2 subsumiert; sonst NIL
;;;; --------------------------------------------------------------------------
(defun variable-subsumes (wert1 bindung1 wert2 bindung2)
  (cond ((and (variable-p wert1) (variable-p wert2))
	 (var-subsume (var-wert wert1 bindung1) bindung1
		      (var-wert wert2 bindung2) bindung2))
	((variable-p wert1)
	 (var-subsume (var-wert wert1 bindung1) bindung1 wert2 bindung2))
	(t (var-subsume wert1 bindung1 (var-wert wert2 bindung2) bindung2))))


;;;; --------------------------------------------------------------------------
;;;; Funktion       VAR-SUBSUMES
;;;;
;;;; Argument(e)  : zwei Werte WERT1, WERT2 und
;;;;                zwei Bindungslisten BINDUNG1, BINDUNG2
;;;; Wert(e)      : T/NIL
;;;; --------------------------------------------------------------------------
(defun var-subsume (wert1 bindung1 wert2 bindung2)
  (cond ((or (variable-p wert1)
	     (equal wert1 wert2)) t)
	((and (listp wert1) (listp wert2))
	 (subsumes1 wert1 bindung1 wert2 bindung2))
	(t nil)))

;;;; --------------------------------------------------------------------------
;;;; Funktion       VAR-WERT
;;;;
;;;; Argument(e)  : eine Variable VAR und 
;;;;                eine Bindungsliste BINDUNG
;;;; Wert(e)      : der Wert von VAR gdw. VAR gebunden ist; sonst VAR
;;;; Beschreibung : sucht in BINDUNG nach einem Wert fuer VAR, der selbst keine
;;;;                gebundene Variable ist
;;;; --------------------------------------------------------------------------
(defun var-wert (var bindung)
  (let ((wert (second (assoc var bindung))))
    (cond ((not wert) var)
	  ((variable-p wert)
	   (var-wert wert bindung))
	  (t wert))))

