;;;;---------------------------------------------------------------------------
;;;;
;;;;                      Generalisierter LR-Recognizer
;;;;                                 nach
;;;;                               M.Tomita
;;;;
;;;;---------------------------------------------------------------------------


;;;; file : tomita.lsp


;;;; A. Daten
;;;; --------

(defvar *syntax*)
(defvar *lexikon*)
(setf *syntax*
      '((S NP VP PP) (det n v prep) S 
	((S NP VP) (S S PP) (NP n) (NP det n) (NP NP PP) 
	 (PP prep NP) (VP v NP))))
(setf *lexikon*
      '((i (n))
	(saw (v))
	(a (det))
	(man (n))
	(with (prep))
	(telescope (n))))

(defvar *action*)
(setf *action*
      '((0 (det (s 3)) (n (s 4)))
	(1 (prep (s 6)) ($ accept))
	(2 (v (s 7)) (prep (s 6)))
	(3 (n (s 10)))
	(4 (v (r 3)) (prep (r 3)) ($ (r 3)))
	(5 (prep (r 2)) ($ (r 2)))
	(6 (det (s 3)) (n (s 4)))
	(7 (det (s 3)) (n (s 4)))
	(8 (prep (r 1)) ($ (r 1)))
	(9 (v (r 5)) (prep (r 5)) ($ (r 5)))
	(10 (v (r 4)) (prep (r 4)) ($ (r 4)))
	(11 (v (r 6)) (prep (r 6) (s 6)) ($ (r 6)))
	(12 (prep (r 7) (s 6)) ($ (r 7)))))

(defvar *goto*)
(setf *goto*
      '((0 (np 2) (s 1))
        (1 (pp 5))
        (2 (pp 9) (vp 8))
        (6 (np 11))
        (7 (np 12))
        (11 (pp 9))
        (12 (pp 9))))


(defvar *stack*)
(defvar *stack-zaehler*)
(defvar eingabe)
(defvar resultat)
(defvar u)
(defvar a)
(defvar r)
(defvar re)
(defvar q)


;;;; B. Selektoren & Praedikate
;;;; --------------------------


; fuer die Parsingtabelle:

;;;; ACTION
;;;; Selektorfunktion fuer den Action-Teil der Parsingtabelle:
;;;; Liefert als Wert den Eintrag des Feldes 'action(zustand, symbol)'
(defmacro action (zustand symbol)
  `(rest (assoc ,symbol (rest (assoc ,zustand *action*)))))


;;;; GOTO
;;;; Selektorfunktion fuer den Goto-Teil der Parsingtabelle:
;;;; Liefert als Wert den Eintrag des Feldes 'goto(zustand, symbol)'
(defmacro goto (zustand symbol)
  `(second (assoc ,symbol (rest (assoc ,zustand *goto*)))))


;;;; STARTZUSTAND
;;;; Liefert den Startzustand des Parsers
(defmacro startzustand ()
  `(caar *action*))
  

; fuer den Stack:

;;;; ETIKETT
;;;; Liefert das Etikett eines Knotens
(defmacro etikett (knoten)
  `(first (aref *stack* ,knoten)))


;;;; TOECHTER
;;;; Liefert die Liste aller Knoten, zu denen vom Knoten
;;;; KNOTEN eine Kante fuehrt
(defmacro toechter (knoten)
  `(second (aref *stack* ,knoten)))


;;;; GENERIERE-KNOTEN
;;;; Erzeugt einen neuen Zustands-/Symbolknoten inkl. Verweis auf alle
;;;; Tochterknoten (optional) mit dem Etikett k
(defmacro generiere-knoten (k &optional nachfolger)
  `(setf *stack-zaehler* (1+ *stack-zaehler*)
	 (aref *stack* *stack-zaehler*) 
	 (list ,k (or ,nachfolger
		      (list (1- *stack-zaehler*))))))


;;;; FINDE-KNOTEN
;;;; Sucht in einer Menge von Knoten K-MENGE nach einem Knoten mit dem
;;;; Etikett NAME
(defun finde-knoten (name k-menge)
  (cond ((endp k-menge) nil)
	((eq name (etikett (first k-menge)))
	 (first k-menge))
	(t (finde-knoten name (rest k-menge)))))


; fuer die Syntax:

;;;; E-REGEL-P
;;;; Das Praedikat testet, ob REGEL eine Tilgungsregel ist oder nicht.
(defmacro e-regel-p (regel)
  `(endp (rest ,regel)))


;;;; ITE-REGEL
;;;; Liefert die Regel mit dem Index I von *syntax*
(defmacro ite-regel (i)
  `(nth (1- ,i) (regeln)))





;;;; C. Hauptfunktionen
;;;; ------------------


;;;; TOMITA_RECOGNIZE
;;;; Hauptfunktion - Die Variablen EINGABE RESULTAT U A R RE Q sind
;;;; special-Variablen.
;;;;   EINGABE  - Liste von Listen lexikalischer Kategorien
;;;;   RESULTAT - boolesche Variable
;;;;   U        - Menge der nach der Verarbeitung des letzten Wortes erzeugten Knoten
;;;;   A        - Menge der aktiven Knoten
;;;;   R, Re, Q - Menge von Eintraegen, die Reduktionsoperationen ohne/mit
;;;;              Tilgungsregeln und Shift-Operationen steuern.
(defun tomita_recognize (satz)
  (progv
      '(  eingabe                  resultat u   a    r  re  q)
      `(,(kategorien-zuweisen satz)  nil   (0) nil  nil nil nil)
    (initialisiere-stack)
    (dotimes (i (length eingabe) resultat)
      (verarbeite_wort))))


;;;; VERARBEITE_WORT
(defun verarbeite_wort ()
  (setq a u  r nil  re nil  q nil)
  (loop
    (when (and (endp a) (endp r) (endp re))
      (return (shift)))
    (cond (a (tabelle_auswerten))
	  (r (reduce))
	  (re (e-reduce)))))


;;;; TABELLE_AUSWERTEN
;;;; Fuer jeden aktiven Knoten und jede dem naechsten Wort zugeordnete
;;;; lexikalische Kategorie werden alle Eintraege des korrespondierenden
;;;; Feldes der Action-Tabelle ausgewertet
(defun tabelle_auswerten ()
  (loop
    (when (endp a) (return))
    (auswerten (pop a) (first eingabe))))

;;;; AUSWERTEN
(defun auswerten (knoten kategorien)
  (dolist (kategorie kategorien)
    (dolist (aktion (action (etikett knoten) kategorie))
      (cond ((eq aktion 'accept) (setq resultat t))
	    ((eq (first aktion) 's)
	     (setq q (neue-shift-zustaende (second aktion)
					   knoten
					   kategorie)))
	    ((and (eq (first aktion) 'r)
		  (not (e-regel-p (ite-regel (second aktion)))))
	     (dolist (x (toechter knoten)) 
	       (push `(,knoten ,x ,(second aktion)) r)))
	    (t (push `(,knoten ,(second aktion)) re))))))



;;;; SHIFT
;;;; Jeder Eintrag in Q hat das Format:
;;;; <neuer-Zustand (alter-Zustand-Index*) lexikalische Kategorie>
(defun shift ()
  (setq u nil)
  (dolist (eintrag q)
    (generiere-knoten (third eintrag) (second eintrag))
    (generiere-knoten (first eintrag))
    (push *stack-zaehler* u))
    (setq q nil
	  eingabe (rest eingabe)))

;;;; NEUE-SHIFT-ZUSTAENDE
;;;; Erzeugt neue Eintraege in Q
(defun neue-shift-zustaende (neu alt kat)
  (let ((eintrag (assoc neu q)))
    (if (and eintrag (eq (third eintrag) kat))
	(cons (list neu
		    (adjoin alt (second eintrag))
		    kat)
	      (remove eintrag q :test #'equal))
        (cons (list neu (list alt) kat) q))))



;;;; E-REDUCE
;;;; Jeder Eintrag in Re hat das Format:
;;;; <neuer-Zustand (alter-Zustand-Index*) Regelindex>
(defun e-reduce ()
  (FORMAT t "Funktion:  REDUCE~%")
  (FORMAT t "R :  ~a~%Q :  ~a~%" r q)
  (setq u nil)
  (dolist (eintrag re)
    (generiere-knoten (third eintrag) (second eintrag))
    (generiere-knoten (first eintrag))
    (push *stack-zaehler* u))
  (setq re nil a u))
    
;;;; NEUE-E-REDUCE-KNOTEN
;;;; Erzeugt in Re neue Eintraege
(defun neue-e-reduce-knoten (alt regelindex)
  (let* ((n (first (ite-regel regelindex)))
	 (neu (goto (etikett alt) n))
	 (eintrag (assoc neu re)))
    (if (and eintrag (eq (third eintrag) n))
	(cons (list neu (adjoin alt (second eintrag)) n)
	      (remove eintrag re :test #'equal))
	(cons (list neu (list alt) n) re))))



;;;; REDUCE
;;;; Jeder Eintrag in R hat das Format:
;;;; <alter-Zustand-Index Symbolknoten-Vorgaenger-Index Regel-Index>
;;;; Es werden folgende lokale Variablen verwendet:
;;;;    N    - linke Seite der zur Reduktion verwendeten Regel
;;;;    K    - letze bei einer Reduktion zu entfernender Symbolknoten
;;;;    PAAR - (Zustand-nach-Reduktion (Zustaende-vor-Reduktion))
;;;;    V    - Knoten aus U
(defun reduce ()
  (let ((n nil))
    (dolist (eintrag r)
      (setq n (first (ite-regel (third eintrag))))
      (dolist (k (reduce-knoten (list (second eintrag))
				(reduce-laenge (third eintrag))))
	(dolist (paar (neue-reduce-zustaende (toechter k) n))
	  (let ((v (finde-knoten (first paar) u)))
	    (cond (v (reduce1 v n paar))
		  (t (generiere-knoten n (second paar))
		     (generiere-knoten (first paar))
		     (setq a (adjoin *stack-zaehler* a)
			   u (adjoin *stack-zaehler* u)))))))))
  (setq r nil))

;;;; REDUCE1
;;;; Verarbeitet den Fall, dass es bereits einen Knoten gibt, dessen
;;;; Etikett mit dem zu generierenden Knoten identisch ist.
(defun reduce1 (v n paar)
  (let ((w (finde-knoten n (toechter v))))
    (unless (and w (equal (toechter w) (second paar)))
      (cond (w (setf (second (aref *stack* w))
		     (union (toechter w) (second paar))))
	    (t (generiere-knoten n (second paar))
	       (setf (second (aref *stack* v))
			     (adjoin *stack-zaehler* (toechter v)))))
      (unless (member v a)
	(dolist (kategorie (first eingabe))
	  (dolist (eintrag (action (first paar) kategorie))
	    (when (and (eq (first eintrag) 'r)
		       (not (e-regel-p (ite-regel (second eintrag)))))
	      (push (list v n (second eintrag)) r))))))))



;;;; D. Hilfsfunktionen
;;;; ------------------

;;;; INITIALISIERE-STACK
;;;; Generiert den Stack und traegt einen dem Startzustand des Parsers
;;;; korrespondierenden Knoten ein.
(defun initialisiere-stack ()
  (setf *stack* (make-array 100 :initial-element nil)
	(aref *stack* 0) (list (startzustand) nil)
	*stack-zaehler* 0))


;;;; KATEGORIEN-ZUWEISEN
;;;; Liefert eine Liste, die fuer jedes Wort aus Satz die Liste aller
;;;; ihm durch das Lexikon zugewiesenen Kategorien enthaelt.
(defun kategorien-zuweisen (satz)
  (let ((resultat nil) (kategorien nil))
    (dolist (wort satz (append resultat '(($))))
      (if (setq kategorien (eintrag-kategorien wort))
	  (setf resultat (append resultat (list kategorien)))
	  (return (format t "~%Fuer das Wort: ~a existiert kein Lexikoneintrag!~%"
			  wort))))))


;;;; REDUCE-KNOTEN
;;;; Liefert die Menge aller Symbolknoten, die bei einer Reduce-Operation
;;;; im letzten Schritt noch entfernt werden muessen.
(defun reduce-knoten (knoten laenge)
  (if (= laenge 0)
      knoten
      (reduce-knoten (berechne-neue-knoten knoten)
		     (- laenge 2))))

;;;; BERECHNE-NEUE-KNOTEN
;;;; Berechnet alle von den Symbolknoten SKNOTEN aus erreichbaren
;;;; Symbolknoten (Nachfolger der von ihnen dominierten Zustandsknoten)
(defun berechne-neue-knoten (sknoten)
  (let ((z-toechter nil) (s-toechter nil))
    (dolist (s sknoten)
      (setq z-toechter (union z-toechter (toechter s))))
    (dolist (z z-toechter s-toechter)
      (setq s-toechter (union s-toechter (toechter z))))))


;;;; REDUCE-LAENGE
;;;; Berechnet die Zahl der zu entfernenden Knoten.
(defun reduce-laenge (i)
  (- (* (length (rest (ite-regel i))) 2) 2))


;;;; NEUE-REDUCE-ZUSTAENDE
;;;; Berechnet eine Liste von Paaren: erstes Element = Zustand Z, zweites
;;;; Element = Liste von Zustaenden, die durch Reduktion zu Z fuehren
(defun neue-reduce-zustaende (knoten symbol)
  (let ((paare nil) (zustand nil))
    (dolist (k knoten paare)
      (if (assoc (setf zustand
		       (goto (etikett k) symbol)) paare)
	  (push k (second (assoc zustand paare)))
	  (when zustand
	        (push `(,zustand (,k)) paare))))))


