;;;;---------------------------------------------------------------------------
;;;;
;;;;
;;;;                              Graphunifikation
;;;;
;;;;
;;;;---------------------------------------------------------------------------



;;;; Datei: unify-7.lsp
;;;; Datum 29.5.92



;;;; A. Makros
;;;; ---------

;;;; --------------------------------------------------------------------------
;;;; Funktion       PFADE
;;;;
;;;; Argument(e)  : ein DAG
;;;; Wert(e)      : die Pfade des DAGs (Typ: Liste)
;;;; --------------------------------------------------------------------------
(defmacro pfade (dag)
  `(first ,dag))


;;;; --------------------------------------------------------------------------
;;;; Funktion       BINDUNGEN
;;;;
;;;; Argument(e)  : ein DAG
;;;; Wert(e)      : die Variablenbindungen des DAGs (Typ: Liste)
;;;; --------------------------------------------------------------------------
(defmacro bindungen (dag)
  `(second ,dag))


;;;; --------------------------------------------------------------------------
;;;; Funktion       ATTRIBUT
;;;;
;;;; Argument(e)  : ein Pfad (Typ: Liste)
;;;; Wert(e)      : erste Element des Pfades (Typ: Atom)
;;;; --------------------------------------------------------------------------
(defmacro attribut (pfad)
  `(first ,pfad))


;;;; --------------------------------------------------------------------------
;;;; Funktion       WERT
;;;;
;;;; Argument(e)  : ein Pfad (Typ: Liste)
;;;; Wert(e)      : zweite Element des Pfades (Typ: Atom v Liste)
;;;; --------------------------------------------------------------------------
(defmacro wert (pfad)
  `(second ,pfad))


;;;; --------------------------------------------------------------------------
;;;; Funktion       ERGAENZE
;;;;
;;;; Argument(e)  : eine Bindungsliste BINDUNGSLISTE
;;;;                und ein Attribut-Wert Paar BINDUNG (beide vom Typ: Liste)
;;;; Wert(e)      : BINDUNGSLISTE wird unter Verwendung von BINDUNG 
;;;;                aktualisiert
;;;; ------------------------------------------------ -------------------------
(defmacro ergaenze (bindungsliste bindung)
  `(if (assoc (first ,bindung) ,bindungsliste)
       (cons ,bindung 
	     (remove-if #'(lambda (x) (eq (first x) (first ,bindung)))
			,bindungsliste))
       (adjoin ,bindung ,bindungsliste :test #'equal)))



;;;; B. Praedikate
;;;; -------------

;;;; --------------------------------------------------------------------------
;;;; Funktion       VARIABLE-P
;;;;
;;;; Argument(e)  : ein Symbol
;;;; Wert(e)      : T/NIL
;;;; Beschreibung : Variablen beginnen mit einem "underscore" (_)
;;;; --------------------------------------------------------------------------
(defun variable-p (symbol)
  (and (symbolp symbol)
       (equal (char (symbol-name symbol) 0)
	      #\_)))


;;;; --------------------------------------------------------------------------
;;;; Funktion       UNBOUND-VARIABLE-P
;;;;
;;;; Argument(e)  : ein Symbol
;;;; Wert(e)      : T/NIL
;;;; Beschreibung : Ein Symbol gilt als ungebunden, wenn fuer es kein Eintrag
;;;;                in der Bindungsliste seines DAgs existiert. 
;;;;                Besonderheit: Wird eine ungebundene Variable mit einer 
;;;;                anderen unifiziert, gilt sie als gebunden.
;;;; --------------------------------------------------------------------------
(defun unbound-variable-p (var bindungen)
       (not (assoc var bindungen)))




;;;; C. Hauptprogramm
;;;; ----------------

;;;; --------------------------------------------------------------------------
;;;; Funktion       UNIFY
;;;;
;;;; Argument(e)  : zwei DAGs: DAG1, DAG2 (Typ: Paar-Listen [Pfade Bindugen])
;;;; Wert(e)      : ein DAG DAG2'/NIL
;;;; Beschreibung : Gelingt die Unifikation beider DAGs, wird der auf der
;;;;                Grundlage von DAG2 gebildete Unifikator als Wert zurueck-
;;;;                gegeben; sonst NIL.
;;;; --------------------------------------------------------------------------
(defun unify (dag1 dag2)
  (unif (pfade dag1)
	(pfade dag2)
	(union (bindungen dag1) (bindungen dag2) :test #'equal)))


;;;; --------------------------------------------------------------------------
;;;; Funktion       UNIF
;;;;
;;;; Argument(e)  : zwei Mengen von Pfaden: DAG1 und DAG2 und 
;;;;                eine Bindungsmenge BINDUNGEN
;;;; Wert(e)      : der Unifikator/NIL
;;;; Beschreibung : Hauptfunktion des Programms
;;;;                Es wird versucht, die beiden DAGs zu unifizieren. 
;;;; --------------------------------------------------------------------------
(defun unif (dag1 dag2 bindungen)
  (let ((eintrag1 (first dag1))
	(eintrag2 (assoc (attribut (first dag1)) dag2)))
  (cond ((not dag1) (list dag2 bindungen))      ; DAG1 leer   
	((not dag2) (list dag1 bindungen))        ; DAG2 leer
	;; Attribut des ersten Pfads von dag1 nicht in dag2 spezifiziert:
	;; Pfad uebertragen
	((not eintrag2)
	 (unif (rest dag1) (cons eintrag1 dag2) bindungen))
	;; Gleiche (atomare) Werte:
	;; unifiziere den Rest von DAG1 mit DAG2
	((equal eintrag1 eintrag2)
	 (unif (rest dag1) dag2 bindungen))
	;; Wert des ersten Pfades in DAG1 bzw. der des zugehoerigen
	;; Pfades in DAG2 ist eine Variable:
	((or (variable-p (wert eintrag1))
	     (variable-p (wert eintrag2)))
	 (verarbeite-variablen dag1 eintrag1 dag2 eintrag2 bindungen))
	;;  Wert des ersten Pfades in DAG1 und der des zugehoerigen
	;; Pfades ind DAG2 ist eine Liste:
	;; unifiziere diese Listen
	((and (listp (wert eintrag1)) (listp (wert eintrag2)))
	 (listen-unif dag1 eintrag1 dag2 eintrag2 bindungen))
	;; Unifikation scheitert wegen inkompatibler Werte
	(t nil))))                             


;;;; --------------------------------------------------------------------------
;;;; Funktion       LISTEN-UNIF
;;;;
;;;; Argument(e)  : zwei Pfadmengen: DAG1, DAG2
;;;;                zwei Pfade aus DAG1, DAG2: EINTRAG1, EINTRAG2 und
;;;;                die Bindungsliste
;;;; Wert(e)      : der Unifikator von DAG1, DAG2  / NIL
;;;; Beschreibung : Wenn die Werte (Typ: Liste) beider Eintraege unifizieren,
;;;;                wird die Unifikation von DAG1, DAG2 fortgesetzt.
;;;; --------------------------------------------------------------------------
(defun listen-unif (dag1 eintrag1 dag2 eintrag2 bindungen)
  (let ((unifikator (unif (wert eintrag1) (wert eintrag2) bindungen)))
    (when unifikator
      (unif (rest dag1)
	     (cons (list (first eintrag2) (first unifikator))
			 (remove eintrag2 dag2 :test #'equal))
	     (second unifikator)))))


;;;; --------------------------------------------------------------------------
;;;; Funktion       VERARBEITE-VARIABLEN
;;;;
;;;; Argument(e)  : zwei Pfadmengen: DAG1, DAG2
;;;;                zwei Pfade: EINTRAG1, EINTRAG2 und
;;;;                die Bindungsmenge BINDUNGEN
;;;; Wert(e)      : der Unifikator fuer DAG1/DAG2 relativ zu BINDUNGSMENGE
;;;;                bzw. NIL
;;;; Beschreibung : diese Funktion versucht zwei Pfade zu unifizieren, von
;;;;                denen mindestens einer als Wert eine Variable enthaelt
;;;; --------------------------------------------------------------------------
(defun verarbeite-variablen (dag1 eintrag1 dag2 eintrag2 bindungen)
  (cond ((and (variable-p (wert eintrag1))
	      (variable-p (wert eintrag2)))
	 (process-var1var2 dag1 (wert eintrag1) dag2 (wert eintrag2)bindungen))
	((variable-p (wert eintrag1))
	 (process-var1 dag1 eintrag1 dag2 eintrag2 bindungen))
	(t (process-var2 dag1 eintrag1 dag2 eintrag2 bindungen))))


;;;; --------------------------------------------------------------------------
;;;; Funktion       PROCESS-VAR1VAR2
;;;;
;;;; Argument(e)  : zwei Pfadmengen: DAG1, DAG2
;;;;                zwei Variablen: VAR1, VAR2 und
;;;;                die Bindungsmenge BINDUNGEN
;;;; Wert(e)      : der Unifikator fuer DAG1/DAG2 relativ zu BINDUNGEN / NIL
;;;; Beschreibung : diese Funktion verarbeitet zwei Pfade, deren Wert eine 
;;;;                Variable ist.
;;;; --------------------------------------------------------------------------
(defun process-var1var2 (dag1 var1 dag2 var2 bindungen)
  (cond ;; beide Variablen sind ungebunden:
        ((and (unbound-variable-p var1 bindungen)
	      (unbound-variable-p var2 bindungen))
	 (unif (rest dag1) dag2 (ergaenze bindungen (list var2 var1))))
	;; VAR1 ist ungebunden
	((unbound-variable-p var1 bindungen)
	 (unif (rest dag1) dag2 (ergaenze bindungen (list var1 var2))))
	;; VAR2 ist ungebunden
	((unbound-variable-p var2 bindungen)
	 (unif (rest dag1) dag2 (ergaenze bindungen (list var2 var1))))
	;; beide Variablen sind gebunden:
	(t (unif-variablen dag1 var1 dag2 var2 bindungen))))



;;;; --------------------------------------------------------------------------
;;;; Funktion       UNIF-VARIABLEN
;;;;
;;;; Argument(e)  : zwei Pfadmengen: DAG1, DAG2
;;;;                zwei Variablen: VAR1, VAR2 und
;;;;                die Bindungsmenge BINDUNGEN
;;;; Wert(e)      : der Unifikator fuer DAG1/DAG2 relativ zu BINDUNGEN / NIL
;;;; Beschreibung : zwei gebundene Variablen werden miteinander unifiziert. 
;;;; --------------------------------------------------------------------------
(defun unif-variablen (dag1 var1 dag2 var2 bindungen)
  (let ((unifikator (unif (list (list var1 (wert (assoc var1 bindungen))))
			  (list (list var1 (wert (assoc var2 bindungen))))
			  bindungen)))
    (when unifikator
      (unif (rest dag1)
	    dag2
	    (append (remove-if #'(lambda (x)
				   (and (or (eq (first x) var1) 
					    (eq (first x) var2))
					(not (variable-p (second x)))))
			       bindungen)
		    (list (list var2 var1)(list var1 (cadaar unifikator))))))))

	

 
;;;; --------------------------------------------------------------------------
;;;; Funktion       PROCESS-VAR1
;;;;
;;;; Argument(e)  : zwei Pfadmengen: DAG1, DAG2
;;;;                zwei Pfade: EINTRAG1, EINTRAG2 und
;;;;                die Bindungsmenge BINDUNGEN
;;;; Wert(e)      : der Unifikator fuer DAG1/DAG2 relativ zu BINDUNGSMENGE
;;;;                bzw. NIL
;;;; Beschreibung : diese Funktion versucht zwei Pfade zu unifizieren, von
;;;;                denen der erste als Wert eine Variable enthaelt
;;;; --------------------------------------------------------------------------

(defun process-var1  (dag1 eintrag1 dag2 eintrag2 bindungen)
  (if (unbound-variable-p (wert eintrag1) bindungen)
      (unif (rest dag1)
	    (cons eintrag1 (remove eintrag2 dag2 :test #'equal))
	    (ergaenze bindungen (list (wert eintrag1) (wert eintrag2))))
      (let ((unifikator
	      (unif (list (list (attribut eintrag1) 
				(wert (assoc (wert eintrag1) bindungen))))
		    (list eintrag2)
		    bindungen)))
	(when unifikator
	  (unif (rest dag1)
		(cons eintrag1 (remove eintrag2 dag2 :test #'equal))
		(progn
		(if (listp (cadaar unifikator))
		    (ergaenze  bindungen 
			       (list (second eintrag1) (cadaar unifikator)))
		      (ergaenze bindungen (first (second unifikator))))))))))


;;;; --------------------------------------------------------------------------
;;;; Funktion       PROCESS-VAR2
;;;;
;;;; Argument(e)  : zwei Pfadmengen: DAG1, DAG2
;;;;                zwei Pfade: EINTRAG1, EINTRAG2 und
;;;;                die Bindungsmenge BINDUNGEN
;;;; Wert(e)      : der Unifikator fuer DAG1/DAG2 relativ zu BINDUNGSMENGE
;;;;                bzw. NIL
;;;; Beschreibung : diese Funktion versucht zwei Pfade zu unifizieren, von
;;;;                denen der zweite als Wert eine Variable enthaelt
;;;; --------------------------------------------------------------------------
(defun process-var2 (dag1 eintrag1 dag2 eintrag2 bindungen)
  (if (unbound-variable-p (wert eintrag2) bindungen)  ; ungebundene Variable
      (unif (rest dag1)
	    dag2 
	    (ergaenze bindungen (list (wert eintrag2) (wert eintrag1))))
      (let ((unifikator 
	      (unif (list eintrag1)
		   (list (list (attribut eintrag2) 
			       (wert (assoc (wert eintrag2) bindungen))))
		   bindungen)))
	(when unifikator
	  (unif (rest dag1)
	        dag2
		(if (listp (cadaar unifikator))
		    (ergaenze bindungen 
			      (list (second eintrag2) (cadaar unifikator)))
		    (ergaenze bindungen (first (second unifikator)))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; --------------------------------------------------------------------------
;;;; Funktion       NEUE-VARIABLE
;;;;
;;;; Argument(e)  : -------------
;;;; Wert(e)      : ein neues Symbol _VAR
;;;; Beschreibung : das generierte Symbol wird zur Repraesentation einer
;;;;                Variablen verwendet; d.h. (VARIABLE-P VAR) ==> T.
;;;; --------------------------------------------------------------------------
(defun neue-variable ()
  (if (equal (machine-type) "Apollo")
      (intern (gensym "_"))
      (gensym "_")))


;;;; --------------------------------------------------------------------------
;;;; Funktion       KOPIERE-REGEL
;;;;
;;;; Argument(e)  : eine Dag: DAG (Liste)
;;;;                
;;;; Wert(e)      : DAG'
;;;; Beschreibung : erzeugt eine Kopie von DAG, in der ein Satz neuer 
;;;;                Variablen verwendet wird
;;;; --------------------------------------------------------------------------
(defun kopiere-regel (regel bindungen)
  (multiple-value-bind (bind vars) (filter (apply #'append regel) bindungen)
    (ersetze-variablen regel bind vars)))


;;;; --------------------------------------------------------------------------
;;;; Funktion       FINDE-VARIABLEN
;;;;
;;;; Argument(e)  : eine Liste PFADE
;;;; Wert(e)      : eine Liste mit allen Variablen, die in PFADE vorkommen:
;;;; --------------------------------------------------------------------------
(defun finde-variablen (pfade &optional vars)
  (let ((result vars) (reste nil))
    (dolist (pfad pfade (if reste (finde-variablen reste result) result))
      (mapcar #'(lambda (x)
		  (if (variable-p x)
		      (setq result (adjoin x result))
		      (when (listp x)
			(setq reste (append reste x)))))
	      pfad))))



;;;; --------------------------------------------------------------------------
;;;; Funktion       ERSETZE-VARIABLEN
;;;;
;;;; Argument(e)  : eine Pfademenge: PFADE (Liste);
;;;;		    eine Bindungsmenge: BINDUNGEN (Liste) und
;;;;                eine Menge von Variablen: VARIABLEN (Liste)
;;;; Wert(e)      : eine Liste der Form: (PFADE' und BINDUNGEN')
;;;; Beschreibung : ersetzt jedes Vorkommen einer Variablen aus VARIABLEN in
;;;;                PFADE und BINDUNGEN durch eine neue Variable
;;;; --------------------------------------------------------------------------
(defun ersetze-variablen (pfade bindungen variablen)
  (let ((pfade1 pfade)
	(bindungen1 bindungen)
	(v nil))
    (dolist (var variablen (list pfade1 bindungen1))
      (setq v          (neue-variable)
	    pfade1     (subst v var pfade1)
	    bindungen1 (subst v var bindungen1)))))


;;;; --------------------------------------------------------------------------
;;;; Funktion       FILTER
;;;;
;;;; Argument(e)  : eine Pfadmenge: PFADE (Liste) und
;;;;                eine Bindungsmenge: BINDUNGEN (Liste)
;;;; Wert(e)      : BINDUNGEN'
;;;; Beschreibung : diese Funktion entfernt alle Eintraege aus der Bindungs-
;;;;                liste, auf die nicht mehr innerhalb von PFADE Bezug 
;;;; 		    genommen wird
;;;; --------------------------------------------------------------------------
(defun filter (pfade bindungen)
  (do* ((variablen (finde-variablen pfade)
		   (set-difference (finde-variablen neue-bindungen) variablen))
	(vars variablen (union vars variablen))
	(neue-bindungen (remove-if-not #'(lambda (x) 
					   (member (first x) variablen)) 
				       bindungen)
			(remove-if-not #'(lambda (x) 
					   (member (first x) variablen)) 
				       bindungen))
	(ergebnis neue-bindungen
		  (union ergebnis neue-bindungen :test #'equal)))
      ((endp variablen) (values ergebnis vars))))


;;;; --------------------------------------------------------------------------
;;;; Funktion       DEBUG-MODE
;;;;
;;;; Argument(e)  : ----------
;;;; Wert(e)      : NIL
;;;; Beschreibung : setzt systemabhaengig Systemvariablen auf geeignete Werte
;;;; --------------------------------------------------------------------------
(defun debug-mode ()
 (if (equal (machine-type) "Apollo")
     (setq *print-level* nil
	    *print-length* nil
	    top-level:*print-level* nil
	    top-level:*print-length* nil)
      (setq *print-level* nil
	    *print-length* nil
	    lisp::*max-step-indentation* 150
	    lisp::*step-print-level* nil
	    lisp::*step-print-length* nil)))
